import pytest  # noqa:F401
import numpy as np

from numba import cuda

from mumott.data_handling import DataContainer
from mumott.core import john_transform_cuda, john_transform_adjoint_cuda

dc = DataContainer('tests/test_half_circle.h5')
gm = dc.geometry
fields = [np.arange(192, dtype=np.float32).reshape(4, 4, 4, 3)]
outs = [np.array([[[[174., 178., 182.],
                   [186., 190., 194.],
                   [198., 202., 206.],
                   [204., 208., 212.]],
                  [[366., 370., 374.],
                   [378., 382., 386.],
                   [390., 394., 398.],
                   [396., 400., 404.]],
                  [[558., 562., 566.],
                   [570., 574., 578.],
                   [582., 586., 590.],
                   [588., 592., 596.]],
                  [[654., 658., 662.],
                   [666., 670., 674.],
                   [678., 682., 686.],
                   [684., 688., 692.]]]])]

out_fields = [np.arange(48)]
projs = [np.arange(48, dtype=np.float32).reshape(1, 4, 4, 3)]


@pytest.mark.parametrize('field,out', [t for t in zip(fields, outs)])
def test_john_transform_cuda(field, out):
    vector_p = np.array([[0, 1., 0.]])
    vector_j = np.array([[1, 0., 0.]])
    vector_k = np.array([[0, 0., 1.]])
    offsets_j = np.zeros((len(gm)))
    offsets_k = np.zeros((len(gm)))
    projections = np.zeros_like(out).astype(np.float32)
    john_transform_cuda(field, projections, vector_p, vector_j, vector_k, offsets_j, offsets_k)
    print(projections)
    assert np.allclose(projections, out)


@pytest.mark.parametrize('proj,out', [t for t in zip(projs, out_fields)])
def test_john_transform_adjoint_cuda(proj, out):
    vector_p = np.array([[0, 1., 0.]])
    vector_j = np.array([[1, 0., 0.]])
    vector_k = np.array([[0, 0., 1.]])
    offsets_j = np.zeros((len(gm)))
    offsets_k = np.zeros((len(gm)))
    field = np.zeros((4, 4, 4, 3,), dtype=np.float32)
    john_transform_adjoint_cuda(field, proj, vector_p, vector_j, vector_k, offsets_j, offsets_k)
    print(field)
    assert np.allclose(field[:, 0].ravel(), out)


@pytest.mark.parametrize('field,out', [t for t in zip(fields, outs)])
def test_device_john_transform_cuda(field, out):
    vector_p = np.array([[0, 1., 0.]])
    vector_j = np.array([[1, 0., 0.]])
    vector_k = np.array([[0, 0., 1.]])
    offsets_j = np.zeros((len(gm)))
    offsets_k = np.zeros((len(gm)))
    projections = cuda.to_device(np.zeros_like(out).astype(np.float32))
    john_transform_cuda(cuda.to_device(field), projections,
                        vector_p, vector_j, vector_k, offsets_j, offsets_k)
    assert np.allclose(projections.copy_to_host(), out)


@pytest.mark.parametrize('proj,out', [t for t in zip(projs, out_fields)])
def test_device_john_transform_adjoint_cuda(proj, out):
    vector_p = np.array([[0, 1., 0.]])
    vector_j = np.array([[1, 0., 0.]])
    vector_k = np.array([[0, 0., 1.]])
    offsets_j = np.zeros((len(gm)))
    offsets_k = np.zeros((len(gm)))
    field = cuda.to_device(np.zeros((4, 4, 4, 3,), dtype=np.float32))
    john_transform_adjoint_cuda(field, cuda.to_device(proj),
                                vector_p, vector_j, vector_k, offsets_j, offsets_k)
    assert np.allclose(field.copy_to_host()[:, 0].ravel(), out)



import pytest
from mumott.core.hashing import list_to_hash

import numpy as np


class myclass:
    a = 0


list_of_lists = [[1, 2, 3],
                 [np.array([0]), dict(a=1, b=(1, 2, 3))],
                 ['stringstring', b'bytestring', np.array(['array', 'of', 'strings'])],
                 [1 + 1j, np.array((1 + 1j)), [np.array((1, 2)), np.array((4, 5))]],
                 [np.sin(np.pi / 5), np.sqrt(2) / 10, np.sqrt(3) * 1e-32],
                 [None, dict(val=None)],
                 [np.array([None])],
                 [myclass()]]

list_of_hashes = ['e64cb9', 'ae9dc0', 'fddfa8', '247c1d', '953a23', 'cade30', None, None]


@pytest.mark.parametrize('inlist, list_hash', [t for t in zip(list_of_lists, list_of_hashes)])
def test_hashing(inlist, list_hash):
    if list_hash is None:
        with pytest.raises(TypeError, match='object'):
            list_to_hash(inlist)
    else:
        my_hash = list_to_hash(inlist)
        assert my_hash[:6] == list_hash



import logging
import os
from typing import Dict, Any, Union, Tuple, List


import numpy as np
from numpy.typing import NDArray
import h5py as h5

logger = logging.getLogger(__name__)


def save_array_like(key: str, val: Union[Tuple, List, NDArray], group):
    val = np.array(val)
    if val.dtype.kind in 'fcbviuS':
        group.create_dataset(key, data=val, shape=val.shape)
    elif val.dtype.kind in 'U':
        val = val.astype(bytes)
        group.create_dataset(key, data=val, shape=val.shape)
    else:
        logger.warning(f'Data type {val.dtype} not supported, entry will be ignored!')


def save_item(key: str, val: Any, group):
    if val is None:
        logger.warning(f'Entry {key} has value None, which is not supported and will be ignored!')
        return
    if isinstance(val, np.ndarray) or type(val) in (list, tuple):
        save_array_like(key, val, group)
    else:
        val = np.array((val,))
        save_array_like(key, val, group)


def save_dict_recursively(inner_dict: Dict[str, Any], group):
    for key, val in inner_dict.items():
        if isinstance(val, dict):
            group.create_group(key)
            save_dict_recursively(val, group[key])
        else:
            save_item(key, val, group)


def dict_to_h5(dict_to_output: Dict[str, Any], filename: str, overwrite: bool = False) -> None:
    """Function for recursively saving a dictionary as an hdf5 file.

    Example
    -------
    The following snippet demonstrates how to save and read an example dictionary using this functionality.
    In this example, the output file will be overwritten in case it already exists.

    >>> from mumott.output_handling import dict_to_h5
    >>> dict_to_h5(dict_to_output=dict(a=5, b='123', c=dict(L=['test'])),
                   filename='my-example-dict.h5', overwrite=True)

    To read the output file (typically at some later point) we use the ``h5py.File`` context.
    Note that this does not restore the original dictionary.

    >>> import h5py
    >>> file = h5py.File('my-example-dict.h5')

    We can loop over the fields in the hdf5 file using the dict-functionality of the ``h5py.File`` class.

    >>> for k, v in file.items():
    ...     print(k, v)
    a <HDF5 dataset "a": shape (1,), type "<i8">
    b <HDF5 dataset "b": shape (1,), type "|S3">
    c <HDF5 group "/c" (1 members)>

    This allows us also to access individual fields

    >>> print(file['a'][0])
    5

    as well as nested data.

    >>> print(file['c/L'][:])
    [b'test']

    Parameters
    ----------
    dict_to_output
        A ``dict`` with supported entry types, including other ``dict``s. File will be recursively
        structured according to the structure of the ``dict``.
    filename
        The name of the file, including the full path, which will be output.
    overwrite
        Whether to overwrite an existing file with name :attr:`filename`. Default is ``False``.

    Notes
    -----
    Supported entry types in :attr:`dict_to_output`:
        ``int``, ``float``, ``complex``
            Saved as array of size ``1``.
        ``bytes``,  ``str``
            ``str`` is cast to ``bytes-like``. Saved as array of size ``1``.
        ``None``
            Ignored.
        ``np.ndarray``
            Provided ``dtype`` is not ``object``, ``datetime`` or ``timedelta,
            hence arrays of ``None`` are not allowed.
        ``list``, ``tuple``
            Provided they can be cast to allowed, i.e. non-ragged ``np.ndarray``.
        ``dict``
            Assuming entries are allowed types, including ``dict``.
    """
    if os.path.exists(filename) and not overwrite:
        raise FileExistsError('File already exists, and overwrite is set to False.')
    with h5.File(filename, 'w') as file_to_save:
        for key, val in dict_to_output.items():
            if isinstance(val, dict):
                file_to_save.create_group(key)
                save_dict_recursively(val, file_to_save[key])
            else:
                save_item(key, val, file_to_save)
    logger.info(f'File {filename} saved successfully!')



# -*- coding: utf-8 -*-

"""
This module provides functionality for loading, accessing, and manipulating
pre-processed data from tensor tomographic experiments.
"""

from .stack import Geometry
from .data_container import DataContainer

__all__ = [
    'Geometry',
    'DataContainer',
]



import pytest  # noqa
import logging

import numpy as np
from mumott import DataContainer

from mumott.methods.basis_sets import SphericalHarmonics
from mumott.methods.projectors import SAXSProjectorNumba
from mumott.methods.functionals import SIGTT
from mumott.optimization.regularizers import Laplacian
from mumott.optimization.loss_functions import SquaredLoss
from mumott.optimization.optimizers import LBFGS

LOGGER = logging.getLogger(__name__)


def test_optimise(caplog):
    caplog.set_level(logging.WARNING)
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    reg = Laplacian()
    lf.add_regularizer('laplacian', reg, 1e-2)
    test_dict = dict(test1=1, disp=False, test2=2, test3=123, maxiter=1)
    opt = LBFGS(lf, **test_dict)
    result = opt.optimize()
    assert 'Unknown option test1' in caplog.text
    assert np.isclose(result['fun'], 5507.07)
    assert np.isclose(result['nit'], 1)


def test_x0(caplog):
    caplog.set_level(logging.WARNING)
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    reg = Laplacian()
    lf.add_regularizer('laplacian', reg, 1e-1)
    test_dict = dict(disp=False, test1=2, test3=123, maxiter=1, x0=np.ones_like(lf.initial_values))
    opt = LBFGS(lf, **test_dict)
    result = opt.optimize()
    assert 'Unknown option test1' in caplog.text
    assert np.isclose(result['fun'], 3252.08)
    assert np.isclose(result['nit'], 1)


def test_hash(caplog):
    reg = Laplacian()
    assert hex(hash(reg))[2:8] == '1df556'
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    assert hex(hash(lf))[2:8] == 'b52d25'
    reg = Laplacian()
    lf.add_regularizer('laplacian', reg, 1e-2)
    assert hex(hash(lf))[2:8] == '130668'
    lf.regularization_weights['laplacian'] = 0.0005
    assert hex(hash(lf))[2:8] == '1fa2d9'


def add_remove_regularizers():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    rf = Laplacian()
    lf.add_regularizer('laplacian', rf, 1e-2)
    assert 'laplacian' in lf.regularizers
    lf.remove_regularizer()
    assert 'laplacian' not in lf.regularizers


def test_str():
    reg = Laplacian()
    s = str(reg)
    assert '1df556' in s


def test_html():
    reg = Laplacian()
    s = reg._repr_html_()
    assert '1df556' in s



""" Container for class Stack. """
import numpy as np
from numpy.typing import NDArray
from .geometry import Geometry, GeometryTuple
from ..core.hashing import list_to_hash


class Frame:
    """Instances of this class contain data and metadata from a single measurement.
    Typically they are appended to a :class:`Stack <mumott.data_handling.stack.Stack>` object.

    Parameters
    ----------
    data
        Data from measurement, structured into 3 dimensions representing
        the two scanning directions and the detector angle.
    diode
        Diode or transmission data from measurement, structured into
        2 dimensions representing the two scanning directions.
    weights
        Weights or masking information, represented as a number
        between ``0`` and ``1``. ``0`` means mask, ``1`` means
        do not mask. Structured the same way as :attr:`data`.
    rotation
        3-by-3 rotation matrix, representing the rotation of the sample in the
        laboratory coordinate system.
    j_offset
        The offset needed to align the projection in the j-direction.
    k_offset
        The offset needed to align the projection in the k-direction.
    """
    def __init__(self,
                 data: NDArray[np.float64] = None,
                 diode: NDArray[np.float64] = None,
                 weights: NDArray[np.float64] = None,
                 rotation: NDArray[np.float64] = np.eye(3, dtype=np.float64),
                 j_offset: np.float64 = np.float64(0),
                 k_offset: np.float64 = np.float64(0)):
        self._key = None
        self._stack = None
        self.data = data
        self.diode = diode
        self.weights = weights
        self.j_offset = j_offset
        self.k_offset = k_offset
        self.rotation = rotation

    @property
    def j_offset(self) -> np.float64:
        """ The offset needed to align the projection in the j-direction."""
        if self._stack is None:
            return self._j_offset
        else:
            k = self._stack.index_by_key(self._key)
            return self._stack.geometry.j_offsets[k]

    @j_offset.setter
    def j_offset(self, value) -> None:
        self._j_offset = value
        if self._stack is not None:
            k = self._stack.index_by_key(self._key)
            self._stack.geometry.j_offsets[k] = value

    @property
    def k_offset(self) -> np.float64:
        """ The offset needed to align the projection in the k-direction."""
        if self._stack is None:
            return self._k_offset
        else:
            k = self._stack.index_by_key(self._key)
            return self._stack.geometry.k_offsets[k]

    @k_offset.setter
    def k_offset(self, value) -> None:
        self._k_offset = value
        if self._stack is not None:
            k = self._stack.index_by_key(self._key)
            self._stack.geometry.k_offsets[k] = value

    @property
    def rotation(self) -> NDArray[np.float64]:
        """ 3-by-3 rotation matrix, representing the rotation of the sample in the
        laboratory coordinate system. """
        if self._stack is None:
            return self._rotation
        else:
            k = self._stack.index_by_key(self._key)
            return self._stack.geometry.rotations[k]

    @rotation.setter
    def rotation(self, value) -> None:
        self._rotation = value
        if self._stack is not None:
            k = self._stack.index_by_key(self._key)
            self._stack.geometry.rotations[k] = value

    @property
    def data(self) -> NDArray:
        """ Scattering data, structured ``(j, k, w)``, where ``j`` is the pixel in the j-direction,
        ``k`` is the pixel in the k-direction, and ``w`` is the detector segment.
        Before the reconstruction, the data should be normalized by the diode.
        This may already have been done prior to loading the data.
        """
        return np.array([]).reshape(0, 0) if self._data is None else self._data

    @data.setter
    def data(self, val) -> None:
        self._data = val

    @property
    def diode(self) -> NDArray[np.float64]:
        """ The diode readout, used to normalize the data. Can be blank if data is already normalized.
        """
        return np.array([]).reshape(0, 0) if self._diode is None else self._diode

    @diode.setter
    def diode(self, val) -> None:
        self._diode = val

    @property
    def weights(self) -> NDArray:
        """ Weights to be applied multiplicatively during optimization. A value of ``0``
        means mask, a value of ``1`` means no weighting, and other values means weighting
        each data point either less (``weights < 1``) or more (``weights > 1``) than a weight of ``1``.
        """
        return np.array([]).reshape(0, 0) if self._weights is None else self._weights

    @weights.setter
    def weights(self, val) -> None:
        self._weights = val

    @property
    def attached(self):
        """ Returns true if frame is attached to a :class:`Stack <Stack>` object. """
        return self._stack is not None

    def attach_to_stack(self, stack, index):
        """ Used to attach the frame to a stack.
        *This method should not be called by users.*
        """
        if self.attached:
            raise ValueError('This frame is already attached to a stack')
        self._stack = stack
        self._key = index

    @property
    def geometry(self) -> GeometryTuple:
        """ Returns geometry information as a named tuple. """
        return GeometryTuple(rotation=self.rotation, j_offset=self.j_offset, k_offset=self.k_offset)

    @geometry.setter
    def geometry(self, value: GeometryTuple) -> None:
        self.rotation = value.rotation
        self.j_offset = value.j_offset
        self.k_offset = value.k_offset

    def detach_from_stack(self):
        """ Used to detach the frame from a stack.
        *This method should not be called by users.*
        """
        k = self._stack.index_by_key(self._key)
        g = self._stack.geometry[k]
        self._rotation = g.rotation
        self._j_offset = g.j_offset
        self._k_offset = g.k_offset
        self._stack = None
        self._key = None

    @property
    def hash_data(self) -> str:
        """ A hash of :attr:`data`."""
        # np.array wrapper in case data is None
        return list_to_hash([np.array(self.data)])

    @property
    def hash_diode(self) -> str:
        """ A sha1 hash of :attr:`diode`."""
        return list_to_hash([np.array(self.diode)])

    @property
    def hash_weights(self) -> str:
        """ A sha1 hash of :attr:`weights`."""
        return list_to_hash([np.array(self.weights)])

    def __str__(self) -> str:
        wdt = 74
        s = []
        s += ['-' * wdt]
        s += ['Frame'.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=4, precision=5, linewidth=60, edgeitems=2):
            s += ['{:18} : {}'.format('hash_data', self.hash_data[:6])]
            s += ['{:18} : {}'.format('hash_diode', self.hash_diode[:6])]
            s += ['{:18} : {}'.format('hash_weights', self.hash_weights[:6])]
            ss = ', '.join([f'{r}' for r in self.rotation])
            s += ['{:18} : {}'.format('rotation', ss)]
            s += ['{:18} : {}'.format('j_offset', self.j_offset)]
            s += ['{:18} : {}'.format('k_offset', self.k_offset)]
        s += ['-' * wdt]
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += ['<h3>Frame</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>']
        s += ['<tbody>']
        with np.printoptions(threshold=4, precision=5, linewidth=40, edgeitems=2):
            s += ['<tr><td style="text-align: left;">data</td>']
            s += [f'<td>{self.data.shape}</td><td>{self.hash_data[:6]} (hash)</td></tr>']
            s += [f'<tr><td style="text-align: left;">diode</td>'
                  f'<td>{self.diode.shape}</td>']
            s += [f'<td>{self.hash_diode[:6]} (hash)</td></tr>']
            s += [f'<tr><td style="text-align: left;">weights</td>'
                  f'<td>{self.weights.shape}</td>']
            s += [f'<td>{self.hash_weights[:6]} (hash)</td></tr>']
            s += [f'<tr><td style="text-align: left;">rotation</td><td>{self.rotation.shape}</td>']
            s += [f'<td>{self.rotation}</td></tr>']
            s += ['<tr><td style="text-align: left;">j_offset</td><td>1</td>']
            s += [f'<td>{self.j_offset}</td></tr>']
            s += ['<tr><td style="text-align: left;">k_offset</td><td>1</td>']
            s += [f'<td>{self.k_offset}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        return '\n'.join(s)


class Stack:
    """Instances of this class contain data, geometry and other pertinent information
    for a series of measurements.
    The individual measurements are stored as :class:`Frame <mumott.data_handling.stack.Frame>` objects.
    The latter are accessible via list-like operations, which enables, for example, iteration over
    measurements but also retrieval of individual measurements by index, in-place modification or deletion.

    The geometry information (i.e., rotations and offsets for each frame)
    is accessible via the :attr:`geometry` attribute.
    Data, diode readouts, and weights can be retrieved as contiguous arrays
    via the properties :attr:`data`, :attr:`diode`, and :attr:`weights`, respectively.

    Example
    -------
    The following code snippet illustrates how individual measurements can be accessed via list operations.
    For demonstration, here, we use default ("empty") frames.
    In practice the individual measurements are read from a data file via the
    :class:`DataContainer <mumott.data_handling.DataContainer>` class, which makes
    them available via the :attr:`DataContainer.stack <mumott.data_handling.DataContainer.stack>` attribute.

    First we create an empty stack

    >>> from mumott.data_handling.stack import Frame, Stack
    >>> stack = Stack()

    Next we create a frame and attach it to the stack.
    In order to be able to distinguish this frame during this example,
    we assign it a :attr:`Frame.j_offset` of ``0.5``.

    >>> frame = Frame(j_offset=0.5)
    >>> stack.append(frame)

    The geometry information can now be accessed via the stack in several different but equivalent ways,
    namely via the original frame object,

    >>> print(frame.j_offset)

    via indexing the stack

    >>> print(stack[0].geometry.j_offset)

    or by indexing the respective geometry property of the stack itself.

    >>> print(stack.geometry.j_offsets[0])

    We can modify the geometry parameters via any of these properties with identical outcome.
    For example,

    >>> stack[0].j_offset = -0.2
    >>> print(frame.j_offset, stack[0].geometry.j_offset, stack.geometry.j_offsets[0])
    -0.2 -0.2 -0.2

    Next consider a situation where several frames are included in the stack.

    >>> stack.append(Frame(j_offset=0.1))
    >>> stack.append(Frame(j_offset=-0.34))
    >>> stack.append(Frame(j_offset=0.23))
    >>> stack.append(Frame(j_offset=0.78))
    >>> print(stack.geometry.j_offsets)
    [-0.2, 0.1, -0.34, 0.23, 0.78]

    The summary of the stack includes hashes for the data, the diode readout, and the weights.
    This allows one to get a quick indication for whether the content of these fields changes.

    >>> print(stack)
    --------------------------------------------------------------------------
                                      Stack
    --------------------------------------------------------------------------
    hash_data          : ...

    We could, for example, decide to remove an invididual frames as we might
    have realized that the data from that measurement was corrupted.

    >>> del stack[1]
    >>> print(stack)
    --------------------------------------------------------------------------
                                      Stack
    --------------------------------------------------------------------------
    hash_data          : ...

    From the output it is readily apparent that the content of the data field
    has changed as a result of this operation.

    Finally, note that we can also loop over the stack, for example, to print the frames.

    >>> for frame in stack:
    >>>     print(frame)
    ...
    """

    def __init__(self) -> None:
        self._frames = []
        self._keys = []
        self._geometry = Geometry()

    def __delitem__(self, k: int) -> None:
        """ Removes a frame from the stack. """
        if abs(k) > len(self) - int(k >= 0):
            raise IndexError(f'Index {k} is out of bounds for Stack of length {len(self)}.')
        self._frames[k].detach_from_stack()
        del self._frames[k]
        del self._geometry[k]
        del self._keys[k]

    def append(self, frame: Frame) -> None:
        """
        Appends a measurement in the form of a :class:`Frame <mumott.data_handling.stack.Frame>` object.
        Once a frame is attached to a stack, the geometry information of the frame will be synchronized
        with the geometry information of the stack (see :attr:`geometry`).

        Parameters
        ----------
        frame
            :class:`Frame <mumott.data_handling.stack.Frame>` object to be appended.
        """
        if frame.attached:
            raise ValueError('The frame is already attached to a stack')
        assert len(self._frames) == len(self._geometry)
        if len(self) == 0:
            self._geometry.projection_shape = np.array(frame.diode.shape)
        elif not np.allclose(self.diode.shape[1:], frame.diode.shape):
            raise ValueError('Appended frame diode must have the same shape as other frames,'
                             f' but its shape is {frame.diode.shape} while other frames'
                             f' have shape {self.diode.shape[1:]}.')
        self._frames.append(frame)
        self._geometry.append(GeometryTuple(rotation=frame.rotation,
                                            j_offset=frame.j_offset,
                                            k_offset=frame.k_offset))

        frame_key = hash(frame)
        self._keys.append(frame_key)
        frame.attach_to_stack(self, frame_key)

    def __setitem__(self, k: int, frame: Frame) -> None:
        """
        This allows each frame of the stack to be safely modified.
        """
        assert len(self._frames) == len(self._geometry)
        if abs(k) > len(self) - int(k >= 0):
            raise IndexError(f'Index {k} is out of bounds for Stack of length {len(self)}.')

        if frame.attached:
            raise ValueError('The frame is already attached to a stack')
        if not np.allclose(self.diode.shape[1:], frame.diode.shape):
            raise ValueError('New frame diode must have the same shape as other frames,'
                             f' but its shape is {frame.diode.shape} while other frames'
                             f' have shape {self.diode.shape[1:]}.')

        # detach and delete previous frame
        del self[k]

        # attach new frame
        self._frames.insert(k, frame)
        self._geometry.insert(k, GeometryTuple(rotation=frame.rotation,
                                               j_offset=frame.j_offset,
                                               k_offset=frame.k_offset))

        frame_key = hash(frame)
        self._keys.insert(k, frame_key)
        frame.attach_to_stack(self, frame_key)

    def insert(self, k: int, frame: Frame) -> None:
        """ Inserts a frame at a particular index, increasing the indices
        of all subsequent frames by 1. """
        assert len(self._frames) == len(self._geometry)
        if abs(k) > len(self) - int(k >= 0):
            raise IndexError(f'Index {k} is out of bounds for stack of length {len(self)}.')

        if frame.attached:
            raise ValueError('The frame is already attached to a stack.')
        if not np.allclose(self.diode.shape[1:], frame.diode.shape):
            raise ValueError('Inserted frame diode must have the same shape as other frames,'
                             f' but its shape is {frame.diode.shape} while other frames'
                             f' have shape {self.diode.shape[1:]}.')

        self._frames.insert(k, frame)
        self._geometry.insert(k, GeometryTuple(rotation=frame.rotation,
                                               j_offset=frame.j_offset,
                                               k_offset=frame.k_offset))
        self._geometry.projection_shape = np.array(frame.diode.shape)
        frame_key = hash(frame)
        self._keys.insert(k, frame_key)
        frame.attach_to_stack(self, frame_key)

    def __getitem__(self, k: int) -> Frame:
        """
        This allows indexing of and iteration over the stack.
        """
        assert len(self._frames) == len(self._geometry)
        if abs(k) > len(self) - round(float(k >= 0)):
            raise IndexError(f'Index {k} is out of bounds for Stack of length {len(self)}.')
        return self._frames[k]

    def __len__(self) -> int:
        return len(self._frames)

    @property
    def data(self) -> NDArray:
        """ Scattering data, structured ``(n, j, k, w)``, where ``n`` is the projection number,
        ``j`` is the pixel in the j-direction, ``k`` is the pixel in the k-direction,
        and ``w`` is the detector segment. Before the reconstruction, this should
        be normalized by the diode. This may already have been done prior to loading the data.
        """
        if len(self) == 0:
            return np.array([]).reshape(0, 0, 0)
        return np.stack([f.data for f in self._frames], axis=0)

    @property
    def diode(self) -> NDArray:
        """ The diode readout, used to normalize the data. Can be blank if data is already normalized.
        The diode value should not be normalized per frame, i.e., it is distinct from the
        transmission value used in standard tomography."""
        if len(self) == 0:
            return np.array([]).reshape(0, 0, 0)
        return np.stack([f.diode for f in self._frames], axis=0)

    @diode.setter
    def diode(self, val) -> None:
        assert len(self) == len(val)
        for i, frame in enumerate(self._frames):
            frame.diode[...] = val[i]

    @property
    def weights(self) -> NDArray:
        """ Weights applied multiplicatively during optimization. A value of ``0``
        means mask, a value of ``1`` means no weighting, and other values means weighting
        each data point either less (``weights < 1``) or more (``weights > 1``) than a weight of ``1``.
        """
        if len(self) == 0:
            return np.array([]).reshape(0, 0, 0)
        return np.stack([f.weights for f in self._frames], axis=0)

    @weights.setter
    def weights(self, val) -> None:
        assert len(self) == len(val)
        for i, frame in enumerate(self._frames):
            frame.weights[...] = val[i]

    def _get_str_representation(self, max_lines: int = 25) -> str:
        """ Retrieves a string representation of the object with the specified
        maximum number of lines.

        Parameters
        ----------
        max_lines
            The maximum number of lines to return.
        """
        s = []
        wdt = 74
        s = []
        s += ['-' * wdt]
        s += ['Stack'.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=3, edgeitems=1, precision=3, linewidth=60):
            s += ['{:18} : {}'.format('hash_data', self.hash_data[:6])]
            s += ['{:18} : {}'.format('hash_diode', self.hash_diode[:6])]
            s += ['{:18} : {}'.format('hash_weights', self.hash_weights[:6])]
            s += ['{:18} : {}'.format('Number of frames', len(self))]
            s += ['{:18} : {}'.format('Number of pixels j', self.diode.shape[1])]
            s += ['{:18} : {}'.format('Number of pixels k', self.diode.shape[2])]
        truncated_s = []
        leave_loop = False
        while not leave_loop:
            line = s.pop(0).split('\n')
            for split_line in line:
                if split_line != '':
                    truncated_s += [split_line]
                if len(truncated_s) > max_lines - 2:
                    if split_line != '...':
                        truncated_s += ['...']
                    if split_line != ('=' * wdt):
                        truncated_s += ['=' * wdt]
                    leave_loop = True
                    break
            if len(s) == 0:
                leave_loop = True
        truncated_s += ['-' * wdt]
        return '\n'.join(truncated_s)

    def __str__(self) -> str:
        return self._get_str_representation()

    @property
    def hash_data(self) -> str:
        """ A hash of :attr:`data`."""
        # np.array wrapper in case data is None
        return list_to_hash([np.array(self.data)])

    @property
    def hash_diode(self) -> str:
        """ A sha1 hash of :attr:`diode`."""
        return list_to_hash([np.array(self.diode)])

    @property
    def hash_weights(self) -> str:
        """ A sha1 hash of :attr:`weights`."""
        return list_to_hash([np.array(self.weights)])

    def _get_html_representation(self, max_lines: int = 25) -> str:
        """ Retrieves an html representation of the object with the specified
        maximum number of lines.

        Parameters
        ----------
        max_lines
            The maximum number of lines to return.
        """
        s = []
        s += ['<h3>Stack</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>']
        s += ['<tbody>']
        with np.printoptions(threshold=3, edgeitems=1, precision=2, linewidth=40):
            s += ['<tr><td style="text-align: left;">data</td>']
            s += [f'<td>{self.data.shape}</td><td>{self.hash_data[:6]} (hash)</td></tr>']
            s += ['<tr><td style="text-align: left;">diode</td>']
            s += [f'<td>{self.diode.shape}</td><td>{self.hash_diode[:6]} (hash)</td></tr>']
            s += ['<tr><td style="text-align: left;">weights</td>']
            s += [f'<td>{self.weights.shape}</td><td>{self.hash_weights[:6]} (hash)</td></tr>']
            s += ['<tr><td style="text-align: left;">Number of pixels j</td>']
            s += ['<td>1</td>']
            s += [f'<td>{self.diode.shape[1]}</td></tr>']
            s += ['<tr><td style="text-align: left;">Number of pixels k</td>']
            s += ['<td>1</td>']
            s += [f'<td>{self.diode.shape[2]}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        truncated_s = []
        line_count = 0
        leave_loop = False
        while not leave_loop:
            line = s.pop(0).split('\n')
            for split_line in line:
                truncated_s += [split_line]
                if '</tr>' in split_line:
                    line_count += 1
                    # Catch if last line had ellipses
                    last_tr = split_line
                if line_count > max_lines - 1:
                    if last_tr != '<tr><td style="text-align: left;">...</td></tr>':
                        truncated_s += ['<tr><td style="text-align: left;">...</td></tr>']
                    truncated_s += ['</tbody>']
                    truncated_s += ['</table>']
                    leave_loop = True
                    break
            if len(s) == 0:
                leave_loop = True
        return '\n'.join(truncated_s)

    def _repr_html_(self) -> str:
        return self._get_html_representation()

    @property
    def geometry(self) -> Geometry:
        """ Contains geometry information for each frame as well
        as information about the geometry of the whole system. """
        return self._geometry

    def index_by_key(self, key):
        """ Returns an index from a key. """
        return self._keys.index(key)



import pytest  # noqa
import numpy as np

from mumott import DataContainer
from mumott.methods.functionals import SIGTT
from mumott.methods.basis_sets import SphericalHarmonics
from mumott.methods.projectors import SAXSProjectorNumba
from mumott import ProbedCoordinates


def test_class_structure():
    dc = DataContainer('tests/test_half_circle.h5')
    gm = dc.geometry
    s = str(gm)
    assert 'Geometry' in s

    pr = SAXSProjectorNumba(gm)
    s = str(pr)
    assert 'SAXSProjectorNumba' in s
    assert pr.is_dirty is False

    test_array = gm.k_direction_0.copy()
    gm.k_direction_0 = test_array
    assert np.allclose(gm.k_direction_0, test_array)

    assert pr.is_dirty is False
    test_array = np.array((0.5, 0.2, 0.1))
    gm.k_direction_0 = test_array

    assert pr.is_dirty is True
    assert np.allclose(gm.k_direction_0, test_array)

    pc = ProbedCoordinates()
    bs = SphericalHarmonics(pc)
    s = str(bs)
    assert 'SphericalHarmonics' in s

    method = SIGTT(dc, bs, pr)
    s = str(method)
    assert 'SIGTT' in s



from abc import ABC, abstractmethod
from typing import Dict

from ... import DataContainer
from ..basis_sets.base_basis_set import BasisSet
from ..projectors.base_projector import Projector


class Functional(ABC):

    """This is the base class from which specific functionals are being derived.
    """

    def __init__(self,
                 data_container: DataContainer,
                 basis_set: BasisSet,
                 projector: Projector):
        pass

    @abstractmethod
    def get_residuals(self) -> Dict:
        pass

    @abstractmethod
    def __str__(self) -> str:
        pass

    @abstractmethod
    def _repr_html_(self) -> str:
        pass



import logging

from typing import Dict, Any

from scipy.optimize import minimize

from ...core.hashing import list_to_hash
from .base_optimizer import Optimizer
from ..loss_functions.base_loss_function import LossFunction


logger = logging.getLogger(__name__)


class LBFGS(Optimizer):
    """This ``Optimizer`` makes the ``L-BFGS-B`` algorithm from `scipy.optimize
    <https://docs.scipy.org/doc/scipy/reference/optimize.html>`_
    available for usage with an ``LossFunction``.

    Parameters
    ----------
    loss_function : LossFunction
        The objective function to be optimized for using this algorithm.
    kwargs : Dict[str, Any]
        Miscellaneous options. See notes for valid entries.

    Notes
    -----
    Valid entries in :attr:`kwargs` are
        x0
            Initial guess for solution vector. Must be the same size as
            :attr:`functional.coefficients`. Defaults to :attr:`loss_function.initial_values`.
        bounds
            Used to set the ``bounds`` of the optimization method, see `scipy.optimize.minimize
            <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
            documentation for details. Defaults to ``None``.
        maxiter
            Maximum number of iterations. Defaults to ``10``.
        disp
            Whether to display output from the optimizer. Defaults to ``True``
        maxcor
            Maximum number of Hessian corrections to the Jacobian. Defaults to ``10``.
        iprint
            If ``disp`` is true, controls output with no output if ``iprint < 0``,
            convergence output only if ``iprint == 0``, iteration-wise output if
            ``0 < iprint <= 99``, and sub-iteration output if ``iprint > 99``.
        maxfun
            Maximum number of function evaluations, including line search evaluations.
            Defaults to ``20``.
        ftol
            Relative change tolerance for objective function. Changes to absolute change tolerance
            if objective function is ``< 1``, see `scipy.optimize.minimize
            <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html>`_
            documentation, which may lead to excessively fast convergence.
            Defaults to ``1e-3``.
        gtol
            Convergence tolerance for gradient. Defaults to ``1e-5``.
    """

    def __init__(self,
                 loss_function: LossFunction,
                 **kwargs: Dict[str, Any]):
        super().__init__(loss_function, **kwargs)
        # This will later be used to reshape the flattened output.
        self._output_shape = None

    def optimize(self) -> Dict:
        """ Executes the optimization using the options stored in this class
        instance. The optimization will continue to run until convergence,
        or until the maximum number of iterations (``'maxiter'``) is exceeded.

        Returns
        -------
            A ``dict`` of optimization results. See `scipy.optimize.OptimizeResult
            <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.OptimizeResult.html>`_
            for details. The entry ``'x'``, which contains the result, will be reshaped using
            the shape of the gradient from :attr:`loss_function`.
        """
        lbfgs_kwargs = dict(x0=self._loss_function.initial_values,
                            bounds=None)
        misc_options = dict(maxiter=10,
                            disp=True,
                            maxcor=10,
                            iprint=1,
                            maxfun=20,
                            ftol=1e-3,
                            gtol=1e-5)

        for k in lbfgs_kwargs:
            if k in dict(self):
                lbfgs_kwargs[k] = self[k]

        for k in misc_options:
            if k in dict(self):
                misc_options[k] = self[k]

        for k in dict(self):
            if k not in lbfgs_kwargs and k not in misc_options:
                logger.warning(f'Unknown option {k}, with value {self[k]}, has been ignored.')
        lbfgs_kwargs['x0'] = lbfgs_kwargs['x0'].ravel()

        def loss_function_wrapper(coefficients):
            d = self._loss_function.get_loss(coefficients, get_gradient=True)
            # Store gradient shape to reshape flattened output.
            if self._output_shape is None:
                self._output_shape = d['gradient'].shape
            return d['loss'], d['gradient'].ravel()

        result = minimize(fun=loss_function_wrapper, **lbfgs_kwargs,
                          jac=True, method='L-BFGS-B', options=misc_options)
        result = dict(result)
        result['x'] = result['x'].reshape(self._output_shape)
        return dict(result)

    def __hash__(self) -> int:
        to_hash = [self._options, hash(self._loss_function)]
        return int(list_to_hash(to_hash), 16)



import pytest  # noqa
import logging

import numpy as np
from mumott import DataContainer

from mumott.methods.basis_sets import SphericalHarmonics
from mumott.methods.projectors import SAXSProjectorNumba
from mumott.methods.functionals import SIGTT
from mumott.optimization.loss_functions import SquaredLoss
from mumott.optimization.optimizers import LBFGS

LOGGER = logging.getLogger(__name__)


def test_dict(caplog):
    caplog.set_level(logging.INFO)
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    test_dict = dict(test1=1, disp=False, test2=2, test3=123)
    opt = LBFGS(lf, **test_dict)
    for key, value in test_dict.items():
        assert np.allclose(opt[key], value)
    d = dict(opt)
    for key, value in test_dict.items():
        assert np.allclose(d[key], value)
    opt['test5'] = 100
    assert opt['test5'] == 100
    print(caplog.text)
    assert 'Key test5 added' in caplog.text


def test_optimise(caplog):
    caplog.set_level(logging.WARNING)
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    test_dict = dict(test1=1, disp=False, test2=2, test3=123, maxiter=1)
    opt = LBFGS(lf, **test_dict)
    result = opt.optimize()
    assert 'Unknown option test1' in caplog.text
    assert np.isclose(result['fun'], 5506.90)
    assert np.isclose(result['nit'], 1)


def test_x0(caplog):
    caplog.set_level(logging.WARNING)
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    test_dict = dict(disp=False, test1=2, test3=123, maxiter=1, x0=np.ones_like(lf.initial_values))
    opt = LBFGS(lf, **test_dict)
    result = opt.optimize()
    assert 'Unknown option test1' in caplog.text
    assert np.isclose(result['fun'], 3240.44)
    assert np.isclose(result['nit'], 1)


def test_hash(caplog):
    caplog.set_level(logging.WARNING)
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    test_dict = dict(test1=1, test2=2, test3=123, maxiter=1)
    opt = LBFGS(lf)
    assert hex(hash(opt))[2:8] == '10e6d1'
    for key, value in test_dict.items():
        opt[key] = value
    assert hex(hash(opt))[2:8] == '19e1c0'


def test_str():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    test_dict = dict(test1=1, test2=2, test3=123, maxiter=1)
    opt = LBFGS(lf, **test_dict)
    s = str(opt)
    assert '19e1c0' in s
    assert 'maxiter' in s
    assert 'test1' in s


def test_html():
    dc = DataContainer('tests/test_half_circle.h5')
    bs = SphericalHarmonics()
    pr = SAXSProjectorNumba(dc.geometry)
    meth = SIGTT(dc, bs, pr)
    lf = SquaredLoss(meth)
    test_dict = dict(test1=1, test2=2, test3=123, maxiter=1)
    opt = LBFGS(lf, **test_dict)
    s = opt._repr_html_()
    assert '19e1c0' in s
    assert 'maxiter' in s
    assert 'test1' in s



import logging
from typing import Dict

import numpy as np
from numpy.typing import NDArray

from ... import DataContainer
from ...core.hashing import list_to_hash
from .base_functional import Functional
from ..basis_sets.base_basis_set import BasisSet
from ..projectors.base_projector import Projector


logger = logging.getLogger(__name__)


class SIGTT(Functional):
    """Method for SIGTT, or Spherical Integral Geometric Tensor Tomography.
    This ``Method`` is an appropriate choice for SAXS tensor tomography, as it relies
    on the small-angle approximation. It relies on inverting the John transform
    (also known as the X-ray transform) of a tensor field (where each tensor is a
    representation of a spherical function) by comparing it to scattering data which
    has been corrected for transmission.

    Parameters
    ----------
    data_container : DataContainer
        Container for the data which is to be reconstructed.
    basis_set : BasisSet
        The basis set used for representing spherical functions.
    projector : Projector
        The type of projector used together with this method.
    integration_samples : Optional, int
        Number of samples to use when calculating the mapping from the
        detector to the sphere. Defaults to ``11``, which is sufficiently dense
        without being computationally onerous. If a different method than
        binning detector pixels is used, e.g., fitting a function and selecting
        points from this function, a different nuimber, such as ``1``, may be
        a more appropriate choice.
    """

    def __init__(self,
                 data_container: DataContainer,
                 basis_set: BasisSet,
                 projector: Projector,
                 integration_samples: int = 11):
        self._data_container = data_container
        self._geometry_hash = hash(data_container.geometry)
        self._basis_set = basis_set
        self._basis_set_hash = hash(self._basis_set)
        self._projector = projector
        # GPU-based projectors need float32
        self._dtype = projector.dtype
        self._coefficients = np.zeros((*self._data_container.geometry.volume_shape,
                                      len(self._basis_set)), dtype=self._dtype)
        self._integration_samples = integration_samples
        # Check if full circle appears covered in data or not.
        delta = np.abs(self._data_container.geometry.detector_angles[0] -
                       self._data_container.geometry.detector_angles[-1] % (2 * np.pi))
        if abs(delta - np.pi) < min(delta, abs(delta - 2 * np.pi)):
            self._full_circle_covered = False
        else:
            logger.warning('The detector angles appear to cover a full circle.'
                           ' Friedel symmetry will be assumed in the calculation.\n'
                           'If this is incorrect, please set the property full_circle_covered'
                           ' to False.')
            self._full_circle_covered = True
        self._basis_set.probed_coordinates.vector = self._get_probed_coordinates()

    def get_residuals(self,
                      get_gradient: bool = False,
                      get_weights: bool = False) -> Dict:
        """ Calculates a residuals and possibly a gradient between
        coefficients projected using the :attr:`BasisSet` and :attr:`Projector`
        attached to this instance.

        Parameters
        ----------
        get_gradient
            Whether to return the gradient. Default is ``False``.
        get_weights
            Whether to return weights. Default is ``False``. If ``True`` along with
            :attr:`get_gradient`, the gradient will be computed with weights.

        Returns
        -------
            A dictionary containing the residuals, and possibly the
            gradient and/or weights. If gradient and/or weights
            are not returned, their value will be ``None``.
        """
        projection = self._basis_set.forward(self._projector.forward(self._coefficients))
        residuals = projection - self._data_container.stack.data
        if get_gradient:
            # todo: consider if more complicated behaviour is useful,
            # e.g. providing function to be applied to weights
            if get_weights:
                gradient = self._projector.adjoint(
                    self._basis_set.gradient(residuals *
                                             self._data_container.weights).astype(self._dtype))
            else:
                gradient = self._projector.adjoint(
                    self._basis_set.gradient(residuals).astype(self._dtype))
        else:
            gradient = None

        if get_weights:
            weights = self._data_container.weights
        else:
            weights = None

        return dict(residuals=residuals, gradient=gradient, weights=weights)

    # Do not simply reuse contents of this method by copying and pasting, refactor instead.
    def _get_probed_coordinates(self) -> NDArray:
        """
        Calculates and returns the probed polar and azimuthal coordinates on the unit sphere at
        each angle of projection and for each detector segment in the system's geometry.
        """
        n_proj = len(self._data_container.geometry)
        n_seg = len(self._data_container.geometry.detector_angles)
        probed_directions_zero_rot = np.zeros((n_seg,
                                               self._integration_samples,
                                               3))
        # Impose symmetry if needed.
        if not self._full_circle_covered:
            shift = np.pi
        else:
            shift = 0
        det_bin_middles_extended = np.copy(self._data_container.geometry.detector_angles)
        det_bin_middles_extended = np.insert(det_bin_middles_extended, 0,
                                             det_bin_middles_extended[-1] + shift)
        det_bin_middles_extended = np.append(det_bin_middles_extended, det_bin_middles_extended[1] + shift)

        for ii in range(n_seg):

            # Check if the interval from the previous to the next bin goes over the -pi +pi discontinuity
            before = det_bin_middles_extended[ii]
            now = det_bin_middles_extended[ii + 1]
            after = det_bin_middles_extended[ii + 2]

            if abs(before - now + 2 * np.pi) < abs(before - now):
                before = before + 2 * np.pi
            elif abs(before - now - 2 * np.pi) < abs(before - now):
                before = before - 2 * np.pi

            if abs(now - after + 2 * np.pi) < abs(now - after):
                after = after - 2 * np.pi
            elif abs(now - after - 2 * np.pi) < abs(now - after):
                after = after + 2 * np.pi

            # Generate a linearly spaced set of angles covering the detector segment
            start = 0.5 * (before + now)
            end = 0.5 * (now + after)
            inc = (end - start) / self._integration_samples
            angles = np.linspace(start + inc / 2, end - inc / 2, self._integration_samples)

            # Make the zero-rotation-frame vectors corresponding to the given angles
            probed_directions_zero_rot[ii, :, :] = np.cos(angles[:, np.newaxis]) * \
                self._data_container.geometry.detector_direction_origin[np.newaxis, :]
            probed_directions_zero_rot[ii, :, :] += np.sin(angles[:, np.newaxis]) * \
                self._data_container.geometry.detector_direction_positive_90[np.newaxis, :]

        # Initialize array for vectors
        probed_direction_vectors = np.zeros((n_proj,
                                             n_seg,
                                             self._integration_samples,
                                             3), dtype=np.float64)
        # Calculate all the rotations
        probed_direction_vectors[...] = \
            np.einsum('kij,mli->kmlj',
                      self._data_container.geometry.rotations_as_array,
                      probed_directions_zero_rot)
        return probed_direction_vectors

    def _update(self, force_update: bool = False) -> None:
        """ Carries out necessary updates if anything changes with respect to
        the geometry or basis set. """
        if not (self.is_dirty or force_update):
            return
        self._dtype = self._projector.dtype
        self._basis_set.probed_coordinates.vector = self._get_probed_coordinates()
        len_diff = len(self._basis_set) - self._coefficients.shape[-1]
        vol_diff = self._data_container.geometry.volume_shape - np.array(self._coefficients.shape[:-1])
        # TODO: Think about whether the ``Method`` should do this or handle it differently
        if np.any(vol_diff != 0) or len_diff != 0:
            logger.warning('Shape of coefficient array has changed, array will be padded'
                           ' or truncated.')
            # save old array, no copy needed
            old_coefficients = self._coefficients
            # initialize new array
            self._coefficients = \
                np.zeros((*self._data_container.geometry.volume_shape, len(self._basis_set)),
                         dtype=self._dtype)
            # for comparison of volume shapes
            shapes = zip(old_coefficients.shape[:-1], self._coefficients.shape[:-1])
            # old coefficients go into middle of new coefficients except in last index
            slice_1 = tuple([slice(max(0, (d-s) // 2), min(d, (s + d) // 2)) for s, d in shapes]) + \
                (slice(0, min(old_coefficients.shape[-1], self._coefficients.shape[-1])),)
            # zip objects are depleted
            shapes = zip(old_coefficients.shape[:-1], self._coefficients.shape[:-1])
            slice_2 = tuple([slice(max(0, (s-d) // 2), min(s, (s + d) // 2)) for s, d in shapes]) + \
                (slice(0, min(old_coefficients.shape[-1], self._coefficients.shape[-1])),)
            # assumption made that old_coefficients[..., 0] correspnds to self._coefficients[..., 0]
            self._coefficients[slice_1] = old_coefficients[slice_2]
            # Assumption may not be true for all representations!
            # TODO: Consider more logic here using e.g. basis set properties.
            if len_diff != 0:
                logger.warning('Size of basis set has changed. Coefficients have'
                               ' been copied over starting at index 0. If coefficients'
                               ' of new size do not line up with the old size,'
                               ' please reinitialize the coefficients.')
        self._geometry_hash = hash(self._data_container.geometry)
        self._basis_set_hash = hash(self._basis_set)

    @property
    def probed_coordinates(self) -> NDArray:
        """ An array of 3-vectors with the (x, y, z)-coordinates
        on the reciprocal space map probed by the method.
        Structured as ``(N, K, I, 3)``, where ``N``
        is the number of projections, ``K`` is the number of
        detector segments, ``I`` is the number of points to be
        integrated over, and the last axis contains the
        (x, y, z)-coordinates.

        Notes
        -----
        The region of the reciprocal space map spanned by
        each detector segment is represented as a parametric curve
        between each segment. This is intended to simulate the effect
        of summing up pixels on a detector screen. For other methods of
        generating data (e.g., by fitting measurements to a curve),
        it may be more appropriate to only include a single point, which
        will then have the same coordinates as the center of a detector
        segments. This can be achieved by setting the property
        :attr:`integration_samples`.
        """
        return self._get_probed_coordinates()

    def __hash__(self) -> int:
        """ Returns a hash of the current state of this instance. """
        to_hash = [self._coefficients,
                   self.probed_coordinates,
                   hash(self._projector),
                   hash(self._data_container.geometry),
                   self._basis_set_hash,
                   self._geometry_hash]
        return int(list_to_hash(to_hash), 16)

    @property
    def coefficients(self) -> NDArray:
        """Optimization coefficients for this method."""
        return self._coefficients

    @coefficients.setter
    def coefficients(self, val: NDArray) -> None:
        self._coefficients = val.reshape(self._coefficients.shape)

    @property
    def is_dirty(self) -> bool:
        """ ``True`` if stored hashes of geometry or basis set objects do
        not match their current hashes. Used to trigger updates """
        return ((self._geometry_hash != hash(self._data_container.geometry)) or
                (self._basis_set_hash != hash(self._basis_set)))

    def __str__(self) -> str:
        wdt = 74
        s = []
        s += ['=' * wdt]
        s += ['SIGTT'.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=4, precision=5, linewidth=60, edgeitems=1):
            s += ['{:18} : {}'.format('BasisSet', self._basis_set.__class__.__name__)]
            s += ['{:18} : {}'.format('Projector', self._projector.__class__.__name__)]
            s += ['{:18} : {}'.format('is_dirty', self.is_dirty)]
            s += ['{:18} : {}'.format('probed_coordinates (hash)',
                                      list_to_hash([self.probed_coordinates])[:6])]
            s += ['{:18} : {}'.format('hash', hex(hash(self))[2:8])]
        s += ['-' * wdt]
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += ['<h3>SIGTT</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>']
        s += ['<tbody>']
        with np.printoptions(threshold=4, edgeitems=2, precision=2, linewidth=40):
            s += ['<tr><td style="text-align: left;">BasisSet</td>']
            s += [f'<td>{1}</td><td>{self._basis_set.__class__.__name__}</td></tr>']
            s += ['<tr><td style="text-align: left;">Projector</td>']
            s += [f'<td>{len(self._projector.__class__.__name__)}</td>'
                  f'<td>{self._projector.__class__.__name__}</td></tr>']
            s += ['<tr><td style="text-align: left;">Is dirty</td>']
            s += [f'<td>{1}</td><td>{self.is_dirty}</td></tr>']
            s += ['<tr><td style="text-align: left;">probed_coordinates</td>']
            s += [f'<td>{self.probed_coordinates.shape}</td>'
                  f'<td>{list_to_hash([self.probed_coordinates])[:6]} (hash)</td></tr>']
            s += ['<tr><td style="text-align: left;">Hash</td>']
            h = hex(hash(self))
            s += [f'<td>{len(h)}</td><td>{h[2:8]}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        return '\n'.join(s)



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Dict, Any

import numpy as np
from numpy.typing import NDArray

from ...core.hashing import list_to_hash
from ...methods.functionals.base_functional import Functional
from ..regularizers.base_regularizer import Regularizer


class LossFunction(ABC):

    """This is the base class from which specific loss functions are derived.

    Parameters
    ----------
    functional
        A class derived from :class:`Functional <mumott.methods.functionals.base_functional.Functional>`
    use_weights
        Whether to multiply residuals with weights before calculating the residual norm. The calculation
        is also applied to the gradient.
    preconditioner
        A preconditioner to be applied to the gradient. Must have the same shape as
        :attr:`functional.coefficients <mumott.methods.functionals.base_functional.Functional.coefficients>`
        or it must be possible to broadcast by multiplication.
    residual_norm_multiplier
        A multiplier that is applied to the residual norm and gradient. Useful in cases where
        a very small or large loss function value changes the optimizer behaviour.
    """

    def __init__(self,
                 functional: Functional,
                 use_weights: bool = False,
                 preconditioner: NDArray[float] = None,
                 residual_norm_multiplier: float = 1):
        self._functional = functional
        self.use_weights = use_weights
        self.preconditioner = preconditioner
        self.residual_norm_multiplier = residual_norm_multiplier
        self._regularizers = {}
        self._regularization_weights = {}

    def get_loss(self,
                 coefficients: NDArray[float] = None,
                 get_gradient: bool = False):
        """Returns loss function value and possibly gradient based on the given :attr:`coefficients`.

        Notes
        -----
        This method simply calls the methods :meth:`get_residual_norm` and :meth:`get_regularization_norm`
        and sums up their respective contributions.

        Parameters
        ----------
        coefficients
            An ``NDArray`` of values of the same shape as :attr:`functional.coefficients
            <mumott.methods.functionals.base_functional.Functional.coefficients>`. Default value
            is ``None``, which leaves it to :meth:`get_residual_norm`
            to handle the choice of coefficients, which in general defaults to using the coefficients
            of the attached :attr:`functional`.
        get_gradient
            If ``True``, returns a ``'gradient'`` of the same shape as :attr:`functional.coefficients
            <mumott.methods.functionals.base_functional.Functional.coefficients>`.
            Otherwise, the entry ``'gradient'`` will be ``None``.

        Returns
        -------
            A dictionary with at least two entries, ``loss`` and ``gradient``.
        """
        residual_norm = self.get_residual_norm(coefficients, get_gradient)
        regularization = self.get_regularization_norm(self._functional.coefficients, get_gradient)

        result = dict(loss=0., gradient=None)
        result['loss'] += residual_norm['residual_norm'] * self.residual_norm_multiplier
        for name in regularization:
            result['loss'] += regularization[name]['regularization_norm'] * self.regularization_weights[name]

        if get_gradient:
            result['gradient'] = residual_norm['gradient'] * self.residual_norm_multiplier
            if self.preconditioner is not None:
                if self._functional.coefficients.shape[:-1] != self.preconditioner.shape[:-1]:
                    raise ValueError('The first three dimensions of the preconditioner must'
                                     ' have the same size as the coefficients of the functional,'
                                     ' and the last index must be the same or 1, but the'
                                     ' functional coefficients have shape'
                                     f' {self._functional.coefficients.shape},'
                                     ' while the preconditioner has shape'
                                     f' {self.preconditioner.shape}!')
                result['gradient'] *= self.preconditioner
            for name in regularization:
                result['gradient'] += regularization[name]['gradient'] * self.regularization_weights[name]
        return result

    def get_residual_norm(self,
                          coefficients: NDArray[float] = None,
                          get_gradient: bool = False) -> Dict:
        """Returns residual norm and possibly gradient based on the attached
        :attr:`functional <mumott.methods.functionals.base_functional.Functional>`.
        If :attr:`coefficients` is given, :attr:`functional.coefficients
        <mumott.methods.functionals.base_functional.Functional.coefficients>` will be updated
        with these new values, otherwise, the residual norm and possibly the gradient
        will just be calculated using the current coefficients.

        Parameters
        ----------
        coefficients
            An ``NDArray`` of values of the same shape as :attr:`functional.coefficients
            <mumott.methods.functionals.base_functional.Functional.coefficients>`.
        get_gradient
            If ``True``, returns a ``'gradient'`` of the same shape as :attr:`functional.coefficients
            <mumott.methods.functionals.base_functional.Functional.coefficients>`.
            Otherwise, the entry ``'gradient'`` will be ``None``.

        Returns
        -------
            A dictionary with at least two entries, ``residual_norm`` and ``gradient``.
        """
        if coefficients is not None:
            self._functional.coefficients = coefficients
        result = self._get_residual_norm_internal(get_gradient)
        return result

    @abstractmethod
    def _get_residual_norm_internal(self, get_gradient: bool) -> Dict:
        """ Method that implements the actual calculation of the residual norm. """
        pass

    @property
    def use_weights(self) -> bool:
        """ Whether to use weights or not in calculating the residual
        and gradient. """
        return self._use_weights

    @use_weights.setter
    def use_weights(self, val: bool) -> None:
        self._use_weights = val

    @property
    def residual_norm_multiplier(self) -> float:
        """ Multiplicative factor by which the residual norm will be scaled. Can be used,
        together with any :attr:`regularization_weights`, to scale the loss function,
        in order to address unexpected behaviour that arises when some optimizers are given
        very small or very large loss functions. """
        return self._residual_norm_multiplier

    @residual_norm_multiplier.setter
    def residual_norm_multiplier(self, val: float) -> None:
        self._residual_norm_multiplier = val

    @property
    def initial_values(self) -> NDArray[float]:
        """ Initial coefficient values for optimizer; defaults to zeros. """
        return np.zeros_like(self._functional.coefficients)

    @property
    def preconditioner(self) -> NDArray[float]:
        """ Preconditioner that is applied to the gradient by multiplication. """
        return self._preconditioner

    @property
    def preconditioner_hash(self) -> str:
        """ Hash of the preconditioner. """
        if self._preconditioner is None:
            return None
        return list_to_hash([self._preconditioner])[:6]

    @preconditioner.setter
    def preconditioner(self, value: NDArray[float]) -> None:
        self._preconditioner = value

    @property
    @abstractmethod
    def _function_as_str(self) -> str:
        """ Should return a string representation of the associated loss function
        of the residual in Python idiom, e.g. 'L(r) = 0.5 * r ** 2' for squared loss. """
        pass

    @property
    @abstractmethod
    def _function_as_tex(self) -> str:
        """ Should return a string representation of the associated loss function
        of the residual in MathJax-renderable TeX, e.g. $L(r) = \frac{r^2}{2}$ for squared loss"""
        pass

    def get_regularization_norm(self,
                                coefficients: NDArray[float],
                                get_gradient: bool = False) -> Dict[str, Dict[str, Any]]:
        """ Returns the regularization norm, and if requested, the gradient, from all
        regularizers attached to this instance, based on the provided :attr`coefficients`.
        If no coefficients are provided, the ones from the attached :attr:`functional` are used.

        Parameters
        ----------
        coefficients
            An ``np.ndarray`` of values of the same shape as :attr:`functional.coefficients
            <mumott.methods.functionals.base_functional.Functional.coefficients>`.
        get_gradient
            Whether to compute and return the gradient. Default is ``False``

        Returns
        -------
            A dictionary with one entry for each regularizer in :attr:`regularizers`, containing
            ``'regularization_norm'`` and ``'gradient'`` as entries.
        """
        regularization = dict()

        if coefficients is None:
            coefficients = self._functional.coefficients

        for name in self._regularizers:
            reg = self._regularizers[name].get_regularization_norm(coefficients=coefficients,
                                                                   get_gradient=get_gradient)
            sub_dict = dict(gradient=None)
            sub_dict['regularization_norm'] = reg['regularization_norm']

            if get_gradient:
                sub_dict['gradient'] = reg['gradient']
            regularization[name] = sub_dict
        return regularization

    def add_regularizer(self,
                        name: str,
                        regularizer: Regularizer,
                        regularization_weight: float) -> None:
        r""" Add a regularizer to the loss function.

        Parameters
        ----------
        name
            Name of the regularizer, to be used as its key.
        regularizer
            The :class:`Regularizer` instance to be attached.
        regularization_weight
            The regularization weight (often denoted :math:`\lambda`),
            by which the residual norm and gradient will be scaled.
        """
        self._regularizers[name] = regularizer
        self._regularization_weights[name] = regularization_weight

    @property
    def regularizers(self) -> Dict[str, Regularizer]:
        """ The dictionary of regularizers appended to this loss function."""
        return self._regularizers

    @property
    def regularization_weights(self) -> Dict[str, float]:
        """ The dictionary of regularization weights appended to this
        loss function. """
        return self._regularization_weights

    def __str__(self) -> str:
        s = []
        wdt = 74
        s += ['=' * wdt]
        s += [self.__class__.__name__.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=4, precision=5, linewidth=60, edgeitems=1):
            s += ['{:18} : {}'.format('Functional', self._functional.__class__.__name__)]
            s += ['{:18} : {}'.format('Uses weights', self.use_weights)]
            s += ['{:18} : {}'.format('preconditioner_hash', self.preconditioner_hash)]
            s += ['{:18} : {}'.format('residual_norm_multiplier',
                                      self._residual_norm_multiplier)]
            s += ['{:18} : {}'.format('Function of residual', self._function_as_str)]
            s += ['{:18} : {}'.format('hash', hex(hash(self))[2:8])]
        s += ['-' * wdt]
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += [f'<h3>{__class__.__name__}</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>']
        s += ['<tbody>']
        with np.printoptions(threshold=4, edgeitems=2, precision=2, linewidth=40):
            s += ['<tr><td style="text-align: left;">Functional</td>']
            s += [f'<td>{1}</td><td>{self._functional.__class__.__name__}</td></tr>']
            s += ['<tr><td style="text-align: left;">use_weights</td>']
            s += [f'<td>{1}</td><td>{self.use_weights}</td></tr>']
            s += ['<tr><td style="text-align: left;">preconditioner_hash</td>']
            s += [f'<td>{1}</td><td>{self.preconditioner_hash}</td></tr>']
            s += ['<tr><td style="text-align: left;">residual_norm_multiplier</td>']
            s += [f'<td>{1}</td><td>{self._residual_norm_multiplier}</td></tr>']
            s += ['<tr><td style="text-align: left;">Function of residual r</td>']
            s += [f'<td>1</td><td>{self._function_as_tex}</td></tr>']
            s += ['<tr><td style="text-align: left;">Hash</td>']
            h = hex(hash(self))
            s += [f'<td>{len(h)}</td><td>{h[2:8]}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        return '\n'.join(s)

    def __hash__(self) -> int:
        to_hash = [hash(self._functional),
                   self.use_weights,
                   self._residual_norm_multiplier,
                   self._preconditioner]
        for r in self.regularizers:
            to_hash.append(hash(self.regularizers[r]))
            to_hash.append(hash(self.regularization_weights[r]))
        return int(list_to_hash(to_hash), 16)



""" Container for class ProjectionFrame. """
from dataclasses import dataclass
from numpy.typing import NDArray


@dataclass
class ProjectionFrame:
    """
    A dataclass for storing "frames" for instances of
    class:`ProjectionViewer <mumott.output_handling.ProjectionViewer>`.

    Attributes
    ----------
    mean
        The mean value of the data.
    std
        The standard deviation of the data.
    angle_color_quad
        A quadruplet of RGBA colors mapping phase angles
        that indicate the orientation of the image.
    """
    mean: NDArray[float]
    std: NDArray[float]
    angle_color_quad: NDArray[float]



import pytest
import numpy as np
from numba import set_num_threads # noqa
from mumott.core.john_transform import (_check_bounds, _update_position, _get_indices, # noqa
                                        _reduce_indices, _initialize_position,
                                        _get_start_and_end_points, _integrate, _back_integrate,
                                        _project_for_one_pixel, _back_project_for_one_pixel,
                                        _john_transform, _john_transform_adjoint, john_transform,
                                        john_transform_adjoint)
from numpy import float64, float32, int32 # noqa


@pytest.mark.parametrize('position, bounds, expected_output', [
    (np.array((10.5, 15.5, 4.1), dtype=np.float64).reshape(1, 3),
     np.array((20, 18, 9), dtype=np.float64), 1),
    (np.array((10.5, 15.5, -4.1), dtype=np.float64).reshape(1, 3),
     np.array((20, 18, 9), dtype=np.float64), 0),
    (np.array((10.5, 15.5, 7), dtype=np.float64).reshape(1, 3),
     np.array((20, 18, 4), dtype=np.float64), 0),
    (np.array(((10.5, 15.5, 4.1), (4, 7, 2)),
     dtype=np.float64).reshape(2, 3), np.array((20, 18, 9), dtype=np.float64), 1)
])
def test_check_bounds(position, bounds, expected_output):
    out = _check_bounds(position, bounds)
    assert out == expected_output


@pytest.mark.parametrize('position, bounds', [
    (np.array(((10.5, 15.5, 4.1), (4, 7, 2)),
     dtype=np.float32).reshape(2, 3), np.array((20, 18, 9), dtype=np.float64)),
    (np.array(((10.5, 15.5, 4.1), (4, 7, 2)),
     dtype=np.float64).reshape(2, 3).T, np.array((20, 18, 9), dtype=np.float64))
])
def test_check_bounds_exception(position, bounds):
    with pytest.raises(TypeError):
        _ = _check_bounds(position, bounds)


@pytest.mark.parametrize('position, steps, projection_vector, expected', [
    (np.array((10.5, 15.5, 4.1), dtype='float64').reshape(1, 3),
     np.int32(10),
     np.array((0.8, -10.2, 2), dtype='float64'),
     np.array([[18.5, -86.5,  24.1]])
     ),
    (np.array(((10.5, 15.5, 4.1), (4, 7, 2)), dtype='float64').reshape(2, 3),
     np.int32(3),
     np.array((0.8, -10.2, 2), dtype='float64'),
     np.array([[12.9, -15.1, 10.1], [6.4, -23.6, 8.]])
     ),
])
def test_update_position(position, steps, projection_vector, expected):
    _update_position(position, steps, projection_vector)
    assert np.allclose(position, expected)


@pytest.mark.parametrize('position, steps, projection_vector, expected_msg', [
    (np.array(((10.5, 15.5, 4.1), (4, 7, 2)), dtype='float32').reshape(2, 3),
     np.int32(3),
     np.array((0.8, -10.2, 2), dtype='float64'),
     'array(float32, 2d, C)'
     ),
    (np.array(((10.5, 15.5, 4.1), (4, 7, 2)), dtype='float64').reshape(2, 3).T,
     np.int32(3),
     np.array((0.8, -10.2, 2), dtype='float64'),
     'array(float64, 2d, F)'
     ),
])
def test_update_position_raises_error(position, steps, projection_vector, expected_msg):
    with pytest.raises(TypeError) as e:
        _update_position(position, steps, projection_vector)
        assert expected_msg in str(e)


def test_get_indices():
    position = np.array((10.5, 15.5, 4.1), dtype=float64).reshape(1, 3)
    steps = int32(4)
    projection_vector = np.array((0.8, -1., 2), dtype=float64)
    volume_indexer = np.array((32, 32, 32), dtype=float64)
    indices = _get_indices(position, projection_vector, volume_indexer, steps)
    assert np.allclose(indices, np.array([[928],
                                          [992],
                                          [1056],
                                          [1088]]))
    position = np.array(((10.5, 15.5, 4.1), (4, 7, 2)), dtype=float64).reshape(2, 3)
    steps = int32(8)
    projection_vector = np.array((1.75, 1.42, 0.01), dtype=float64)
    volume_indexer = np.array((64, 32, 17), dtype=float64)
    indices = _get_indices(position, projection_vector, volume_indexer, steps)
    assert np.allclose(indices, np.array([[1188,  514],
                                          [1348,  610],
                                          [1540,  770],
                                          [1636,  962],
                                          [1828, 1122],
                                          [1988, 1250],
                                          [2180, 1410],
                                          [2276, 1570]], dtype=int32))
    with pytest.raises(TypeError) as e:
        _ = _get_indices(position.astype(float32), projection_vector, volume_indexer, steps)
        assert 'array(float32, 2d, C)' in str(e)
    with pytest.raises(TypeError) as e:
        _ = _get_indices(position.T, projection_vector, volume_indexer, steps)
        assert 'array(float64, 2d, F)' in str(e)


def test_reduce_indices():
    indices = np.array([[928, 928, 928, 960, 960, 992, 992,
                         992, 1024, 1024, 1024, 1024]], dtype=int32).reshape(-1, 1)
    sampling_kernel = np.array((0.5,), dtype=float64)
    reduced_indices = np.empty_like(indices[:, 0])
    weights = np.empty_like(indices[:, 0], dtype=float64)
    counter = _reduce_indices(indices, sampling_kernel, reduced_indices, weights)
    assert counter == 4
    assert np.allclose(reduced_indices[:counter], np.array([928, 960, 992, 1024]))
    assert np.allclose(weights[:counter], np.array([1.5, 1., 1.5, 2.]))


def test_initialize_position():
    j = int32(23)
    k = int32(16)
    offsets = np.array(((0.3, -0.8), (-0.4, 0.1)), dtype=float64)
    projection_bounds = np.array((18, 20), dtype=float64)
    unit_vector_p = np.array((0.5, 1.2, 0.9), dtype=np.float64)
    unit_vector_p = unit_vector_p / np.linalg.norm(unit_vector_p)
    unit_vector_j = np.array((-1.2, 0.5, 0), dtype=np.float64)
    unit_vector_j = unit_vector_j - unit_vector_p * np.dot(unit_vector_j, unit_vector_p)
    unit_vector_j = unit_vector_j / np.linalg.norm(unit_vector_j)
    unit_vector_k = np.cross(unit_vector_j, unit_vector_p)
    scalar_bound = -60.
    volume_shift = np.array((18, 18, 18), dtype=np.float64)
    position = np.zeros((2, 3), dtype=np.float64)
    _initialize_position(j, k, offsets, projection_bounds, unit_vector_p, unit_vector_j,
                         unit_vector_k, scalar_bound, volume_shift, position)
    assert np.allclose(position, np.array([[-31.65501972,   7.50795268, -45.83373685],
                                           [-30.53017164,   7.34758806, -46.24483294]]))


def test_get_start_and_end_points():
    position = np.array((-0.5, -2, -1), dtype=np.float64).reshape(1, 3)
    projection_vector = np.array((0.3, 1.2, 0.6), dtype=np.float64)
    volume_bounds = np.array((10, 15, 18), dtype=np.float64)
    steps = _get_start_and_end_points(position, projection_vector, volume_bounds)
    assert steps == 12
    assert np.allclose(position, np.array([[0.1, 0.4, 0.2]]))
    position = np.array((-0.5, -2, -1), dtype=np.float64).reshape(1, 3)
    projection_vector = np.array((-0.3, -1.2, -0.6), dtype=np.float64)
    steps = _get_start_and_end_points(position, projection_vector, volume_bounds)
    assert steps == 0
    position = np.array((-0.5, -2, -1), dtype=np.float64).reshape(1, 3)
    projection_vector = np.array((1e-16, 1e-16, 1e-16), dtype=np.float64)
    steps = _get_start_and_end_points(position, projection_vector, volume_bounds)
    assert steps == 0
    position = np.array((0.5, 2, 1), dtype=np.float64).reshape(1, 3)
    projection_vector = np.array((0.3, 1.2, 0.6), dtype=np.float64)
    volume_bounds = np.array((-10, -15, -18), dtype=np.float64)
    steps = _get_start_and_end_points(position, projection_vector, volume_bounds)
    assert steps == 0


def test_integrate():
    position = np.array(((0.1, 0.2, 1.1), (0.1, 0.2, 1.1)), dtype=float64).reshape(-1, 3)
    projection_vector = np.array((0.8, 0.61, 1.), dtype=float64)
    volume_indexer = np.array((4, 4, 4), dtype=float64)
    volume = np.ones((4, 4, 4, 2), dtype=float64)
    volume[..., 1] *= 2.6
    volume[:, 0, 1, :] *= 1.2
    volume = volume.reshape(-1, 2)
    sampling_kernel = np.array(((0.26, 0.12)), dtype=float64)
    total_steps = int32(3)
    pixel = np.zeros(2, dtype=float64)
    _integrate(position, sampling_kernel, volume, volume_indexer, projection_vector,
               total_steps, pixel)
    assert np.allclose(pixel, np.array([1.14, 2.964], dtype=float64))


def test_back_integrate():
    position = np.array(((0.1, 0.2, 0.1), (0.1, 0.2, 0.15)), dtype=float64).reshape(-1, 3)
    projection_vector = np.array((0.3, 0.21, 0.3), dtype=float64)
    volume_indexer = np.array((2, 2, 2), dtype=float64)
    volume = np.zeros((2, 2, 2, 2), dtype=float64)
    volume = volume.reshape(-1, 2)
    sampling_kernel = np.array(((0.26, 0.12)), dtype=float64)
    total_steps = int32(3)
    distributor = np.array((0.9, 1.7), dtype=float64)
    _back_integrate(position, distributor, sampling_kernel, volume,
                    volume_indexer, projection_vector,
                    total_steps)
    assert np.allclose(volume, np.array([[1.026, 1.938],
                                         [0., 0.],
                                         [0., 0.],
                                         [0., 0.],
                                         [0., 0.],
                                         [0., 0.],
                                         [0., 0.],
                                         [0., 0.]]))


def test_project_for_one_pixel():
    position = np.array(((-0.1, -0.2, -0.1), (-0.1, -0.2, -0.15)), dtype=float64).reshape(-1, 3)
    projection_vector = np.array((0.3, 0.3, 0.3), dtype=float64)
    volume_indexer = np.array((4, 4, 4), dtype=float64)
    volume_bounds = np.array((4, 4, 4), dtype=float64)
    volume = np.ones((4, 4, 4, 2), dtype=float64)
    volume[..., 1] *= 6.3
    volume[:, 0, 2, :] *= 2.2
    volume = volume.reshape(-1, 2)
    sampling_kernel = np.array(((0.26, 0.12)), dtype=float64)
    unit_vector_j = np.array((-1.2, 0.5, 0), dtype=np.float64)
    unit_vector_k = np.array((1.2, -0.5, 0), dtype=np.float64)
    pixel = np.zeros(2, dtype=float64)
    _project_for_one_pixel(position, projection_vector, unit_vector_j, unit_vector_k,
                           sampling_kernel, volume_bounds, pixel, volume, volume_indexer)
    assert np.allclose(pixel, np.array([4.56, 28.728]))


def test_back_project_for_one_pixel():
    position = np.array(((-0.1, -0.2, -0.1), (-0.1, -0.2, -0.15)), dtype=float64).reshape(-1, 3)
    projection_vector = np.array((0.1, 0.2, 0.1), dtype=float64)
    volume_indexer = np.array((2, 2, 2), dtype=float64)
    volume_bounds = np.array((2, 2, 2), dtype=float64)
    volume = np.zeros((2, 2, 2, 2), dtype=float64)
    volume = volume.reshape(-1, 2)
    distributor = np.array((0.9, 1.7), dtype=float64)
    sampling_kernel = np.array(((0.26, 0.12)), dtype=float64)
    unit_vector_j = np.array((-1.2, 0.5, 0), dtype=np.float64)
    unit_vector_k = np.array((1.2, -0.5, 0), dtype=np.float64)
    _back_project_for_one_pixel(position,  distributor, sampling_kernel,
                                projection_vector, unit_vector_j, unit_vector_k,
                                volume_bounds, volume, volume_indexer)
    assert np.allclose(volume, np.array([[1.368, 2.584],
                                         [0., 0.],
                                         [1.368, 2.584],
                                         [0., 0.],
                                         [0., 0.],
                                         [0., 0.],
                                         [0., 0.],
                                         [0., 0.]]))


def test_john_transform():
    projection = np.zeros((2, 2, 2), dtype=float64)
    projection_offsets = np.array((-0.1, 0.1), dtype=float64)
    volume = np.ones((8, 2), dtype=float64) * 2.4
    volume[3, 0] = 1.6
    volume[5, 1] = 9.12
    volume_bounds = np.array((2, 2, 2), dtype=float64)
    volume_indexer = np.array((4, 2, 1), dtype=float64)
    unit_vector_p = np.array((0.5, 1.2, 0.9), dtype=np.float64)
    unit_vector_p = unit_vector_p / np.linalg.norm(unit_vector_p)
    unit_vector_j = np.array((-1.2, 0.5, 0), dtype=np.float64)
    unit_vector_j = unit_vector_j - unit_vector_p * np.dot(unit_vector_j, unit_vector_p)
    unit_vector_j = unit_vector_j / np.linalg.norm(unit_vector_j)
    unit_vector_k = np.cross(unit_vector_j, unit_vector_p)
    sampling_kernel = np.array((0.15, 0.7, 0.15,), dtype=float64)
    kernel_offsets = np.array(((-0.15, 0.), (0., -0.1), (0.15, 0.1)), dtype=float64).T
    kernel_offsets = np.ascontiguousarray(kernel_offsets)
    step_size = float64(0.5)
    set_num_threads(1)
    scalar_bound = float64(-3.5)
    _john_transform(projection, projection_offsets,
                    volume, volume_bounds,
                    volume_indexer, unit_vector_p,
                    unit_vector_j, unit_vector_k,
                    sampling_kernel, kernel_offsets,
                    step_size, scalar_bound)
    assert np.allclose(projection, np.array([[[3.6, 12.672],
                                              [2.4, 2.4]],
                                             [[4.34, 4.8],
                                              [2.4, 2.4]]]))


def test_john_transform_adjoint():
    projection = np.ones((2, 2, 2), dtype=float64)
    projection[:, 1, :] *= 1.5
    projection[1, :, 0] *= 2.7
    projection[0, :, :] *= 9.9
    projection_offsets = np.array((0.1, -0.1), dtype=float64)
    volume = np.zeros((1, 8, 2), dtype=float64)
    volume_bounds = np.array((2, 2, 2), dtype=float64)
    volume_indexer = np.array((4, 2, 1), dtype=float64)
    unit_vector_p = np.array((0.5, 1.2, 0.9), dtype=np.float64)
    unit_vector_p = unit_vector_p / np.linalg.norm(unit_vector_p)
    unit_vector_j = np.array((-1.2, 0.5, 0), dtype=np.float64)
    unit_vector_j = unit_vector_j - unit_vector_p * np.dot(unit_vector_j, unit_vector_p)
    unit_vector_j = unit_vector_j / np.linalg.norm(unit_vector_j)
    unit_vector_k = np.cross(unit_vector_j, unit_vector_p)
    sampling_kernel = np.array((0.15, 0.7, 0.15,), dtype=float64)
    kernel_offsets = np.array(((-0.01, 0.), (0., -0.01), (0.01, 0.01)), dtype=float64).T
    kernel_offsets = np.ascontiguousarray(kernel_offsets)
    step_size = float64(0.5)
    set_num_threads(1)
    scalar_bound = float64(-3.5)
    _john_transform_adjoint(projection, projection_offsets,
                            volume, volume_bounds,
                            volume_indexer, unit_vector_p,
                            unit_vector_j, unit_vector_k,
                            sampling_kernel, kernel_offsets,
                            step_size, scalar_bound)
    assert np.allclose(volume, np.array([[[2.025, 0.75],
                                          [2.7, 1.],
                                          [4.05, 1.5],
                                          [0., 0.],
                                          [14.85, 14.85],
                                          [9.9, 9.9],
                                          [14.85, 14.85],
                                          [0., 0.]]]))


def test_john_transform_2():
    projection = np.zeros((2, 2, 2), dtype=float64)
    projection_offsets = np.array((0.02, 0.05), dtype=float64)
    volume = np.ones((3, 3, 3, 2), dtype=float64) * 2.4
    volume[0, 2, 1, 0] = 1.6
    volume[2, 1, 1, 1] = 9.12
    unit_vector_p = np.array((0.5, 1.2, 0.9), dtype=np.float64)
    unit_vector_p = unit_vector_p / np.linalg.norm(unit_vector_p)
    unit_vector_j = np.array((-1.2, 0.5, 0), dtype=np.float64)
    unit_vector_j = unit_vector_j - unit_vector_p * np.dot(unit_vector_j, unit_vector_p)
    unit_vector_j = unit_vector_j / np.linalg.norm(unit_vector_j)
    unit_vector_k = np.cross(unit_vector_j, unit_vector_p)
    sampling_kernel = np.array((0.15, 0.7, 0.15,), dtype=float64)
    kernel_offsets = np.array(((0.001, 0.), (0., 0.04), (0.002, 0.03)), dtype=float64).T
    kernel_offsets = np.ascontiguousarray(kernel_offsets)
    step_size = float64(0.5)
    set_num_threads(1)
    john_transform(projection,
                   volume, unit_vector_p,
                   unit_vector_j, unit_vector_k,
                   step_size,
                   projection_offsets,
                   sampling_kernel, kernel_offsets)
    assert np.allclose(projection, np.array([[[7.2, 7.2],
                                              [8.4, 11.76]],
                                             [[8.4, 8.4],
                                              [7.2, 7.2]]]))


def test_john_transform_adjoint_2():
    projection = np.ones((2, 2, 2), dtype=float64)
    projection[1, :, 0] *= 0.2
    projection[:, 1, :] *= 3.5
    projection[:, :, 1] *= 2.1
    projection_offsets = np.array((0.02, 0.05), dtype=float64)
    volume = np.zeros((1, 2, 2, 2, 2), dtype=float64)
    unit_vector_p = np.array((0.5, 1.2, 0.9), dtype=np.float64)
    unit_vector_p = unit_vector_p / np.linalg.norm(unit_vector_p)
    unit_vector_j = np.array((-1.2, 0.5, 0), dtype=np.float64)
    unit_vector_j = unit_vector_j - unit_vector_p * np.dot(unit_vector_j, unit_vector_p)
    unit_vector_j = unit_vector_j / np.linalg.norm(unit_vector_j)
    unit_vector_k = np.cross(unit_vector_j, unit_vector_p)
    sampling_kernel = np.array((0.15, 0.7, 0.15,), dtype=float64)
    kernel_offsets = np.array(((0.001, 0.), (0., 0.04), (0.002, 0.03)), dtype=float64).T
    kernel_offsets = np.ascontiguousarray(kernel_offsets)
    step_size = float64(0.5)
    set_num_threads(1)
    john_transform_adjoint(projection,
                           volume, unit_vector_p,
                           unit_vector_j, unit_vector_k,
                           step_size,
                           projection_offsets,
                           sampling_kernel, kernel_offsets)
    assert np.allclose(volume, np.array([[[[[0.17, 1.785],
                                            [0.13, 1.365]],
                                           [[0.7, 7.35],
                                            [0.1, 1.05]]],
                                          [[[2.1, 4.41],
                                            [1.15, 2.415]],
                                           [[3.5, 7.35],
                                            [0., 0.]]]]]))



import pytest # noqa
import numpy as np
from mumott.data_handling import DataContainer
from mumott.data_handling.stack import Stack, Frame
from mumott.data_handling.geometry import GeometryTuple


def test_hashes():
    dc = DataContainer('tests/test_half_circle.h5')
    assert dc.stack[0].hash_data.startswith('ab1de1')
    assert dc.stack[0].hash_diode.startswith('dcf160')
    assert dc.stack[0].hash_weights.startswith('65e22e')
    assert dc.stack.hash_data.startswith('ab1de1')
    assert dc.stack.hash_diode.startswith('dcf160')
    assert dc.stack.hash_weights.startswith('65e22e')
    dc = DataContainer('tests/test_full_circle.h5')
    assert dc.stack[0].hash_data.startswith('475981')
    assert dc.stack[0].hash_diode.startswith('40a75d')
    assert dc.stack[0].hash_weights.startswith('a213c4')
    assert dc.stack.hash_data.startswith('475981')
    assert dc.stack.hash_diode.startswith('40a75d')
    assert dc.stack.hash_weights.startswith('a213c4')


def test_empty_hashes():
    s = Stack()
    assert s.hash_data.startswith('786a02')
    assert s.hash_diode.startswith('786a02')
    assert s.hash_weights.startswith('786a02')
    f = Frame()
    assert f.hash_data.startswith('786a02')
    assert f.hash_diode.startswith('786a02')
    assert f.hash_weights.startswith('786a02')
    s.append(f)
    assert s.hash_data.startswith('786a02')
    assert s.hash_diode.startswith('786a02')
    assert s.hash_weights.startswith('786a02')
    assert s[0].hash_data.startswith('786a02')
    assert s[0].hash_diode.startswith('786a02')
    assert s[0].hash_weights.startswith('786a02')


def test_append():
    stack = Stack()
    for i in range(5):
        f = Frame(diode=np.array((0.5, 1.5)).reshape(1, 2), j_offset=i)
        stack.append(f)
    assert np.allclose(stack[0].diode, (0.5, 1.5))
    f = Frame(diode=np.array((0.5, 1.5)).reshape(2, 1))

    with pytest.raises(ValueError, match='Appended'):
        stack.append(f)
    with pytest.raises(ValueError, match='New'):
        stack[0] = f
    with pytest.raises(ValueError, match='Inserted'):
        stack.insert(0, f)


def test_insert():
    stack = Stack()
    for i in range(5):
        f = Frame(j_offset=i)
        stack.append(f)
    f = Frame(j_offset=5)
    stack.insert(1, f)
    assert len(stack) == 6
    assert stack[1] == f
    assert stack[1].j_offset == 5
    assert stack[5].j_offset == 4


def test_setitem():
    stack = Stack()
    for i in range(5):
        f = Frame(j_offset=i)
        stack.append(f)
    f = Frame(j_offset=5)
    stack[1] = f
    assert len(stack) == 5
    assert stack[1] == f
    assert stack[1].j_offset == 5
    assert stack[4].j_offset == 4


def test_attached():
    stack = Stack()
    for i in range(5):
        f = Frame(j_offset=i)
        stack.append(f)
    other_stack = Stack()
    with pytest.raises(ValueError, match='attached'):
        other_stack.append(f)
    other_stack.append(Frame())

    with pytest.raises(ValueError, match='attached'):
        other_stack[0] = f

    with pytest.raises(ValueError, match='attached'):
        other_stack.insert(0, f)

    del stack[4]
    other_stack.append(f)


def test_size():
    stack = Stack()
    for i in range(5):
        f = Frame(j_offset=i)
        stack.append(f)
    with pytest.raises(IndexError, match='bounds'):
        f = stack[5]
    with pytest.raises(IndexError, match='bounds'):
        f = stack[-6]


def test_str():
    dc = DataContainer('tests/test_half_circle.h5')
    stack = dc.stack
    s = stack._get_str_representation(5)
    assert s.count('\n') <= 6


def test_html():
    dc = DataContainer('tests/test_half_circle.h5')
    stack = dc.stack
    s = stack._get_html_representation(5)
    assert s.count('</tr>') <= 6


def test_set_frame_geo():
    f = Frame()
    g = GeometryTuple(rotation=np.ones(3), j_offset=0, k_offset=5)
    f.geometry = g
    assert np.allclose(f.rotation, g.rotation)
    assert np.allclose(f.j_offset, g.j_offset)
    assert np.allclose(f.k_offset, g.k_offset)



import pytest
import numpy as np

from numba import cuda

from mumott.methods.projectors.saxs_projector_cuda import SAXSProjectorCUDA
from mumott.data_handling import DataContainer
from mumott.data_handling.geometry import GeometryTuple

dc = DataContainer('tests/test_half_circle.h5')
gm = dc.geometry
gm.rotations[0] = np.eye(3)

gm_tuples = [GeometryTuple(rotation=np.eye(3), j_offset=0.2, k_offset=0.5),
             GeometryTuple(rotation=np.eye(3)[[2, 0, 1]], j_offset=0.2453, k_offset=0.5222),
             GeometryTuple(rotation=np.eye(3)[[1, 0, 2]], j_offset=-0.2453, k_offset=0.2)]

gm_hash_table = ['277106', '902494', '161751']

kernel_hash = ['222572', '167889', '185810']

fields = [np.arange(192, dtype=np.float32).reshape(4, 4, 4, 3)]
test_projs = [np.array([[[[174., 178., 182.],
                         [186., 190., 194.],
                         [198., 202., 206.],
                         [204., 208., 212.]],
                        [[366., 370., 374.],
                         [378., 382., 386.],
                         [390., 394., 398.],
                         [396., 400., 404.]],
                        [[558., 562., 566.],
                         [570., 574., 578.],
                         [582., 586., 590.],
                         [588., 592., 596.]],
                        [[654., 658., 662.],
                         [666., 670., 674.],
                         [678., 682., 686.],
                         [684., 688., 692.]]]])]

adjs = [np.arange(48)]
projs = [np.arange(48, dtype=np.float32).reshape(1, 4, 4, 3)]


@pytest.mark.parametrize('gm_tuple, expected_value', [(g, h) for g, h in zip(gm_tuples, gm_hash_table)])
def test_gm_hash(gm_tuple, expected_value):
    dc = DataContainer('tests/test_half_circle.h5')
    geom = dc.geometry
    pr = SAXSProjectorCUDA(geom)
    geom[0] = gm_tuple
    assert pr.is_dirty
    pr._update()
    assert expected_value == str(pr._geometry_hash)[:6]


@pytest.mark.parametrize('field, proj', [f for f in zip(fields, test_projs)])
def test_forward(field, proj):
    pr = SAXSProjectorCUDA(gm)
    p = pr.forward(field)
    assert np.allclose(proj, p)


@pytest.mark.parametrize('field, proj', [f for f in zip(fields, test_projs)])
def test_forward_subset(field, proj):
    dc = DataContainer('tests/test_half_circle.h5')
    geo = dc.geometry
    geo.rotations[0] = np.eye(3)
    geo.append(geo[0])
    pr = SAXSProjectorCUDA(geo)
    p = pr.forward(field, indices=0)
    assert np.allclose(proj, p)
    with pytest.raises(TypeError, match='integer kind'):
        pr.forward(field, indices=np.array('abc'))


@pytest.mark.parametrize('proj, adj', [f for f in zip(projs, adjs)])
def test_adj(proj, adj):
    pr = SAXSProjectorCUDA(gm)
    a = pr.adjoint(proj)
    assert np.allclose(adj, a[:, 0].ravel())


@pytest.mark.parametrize('field, proj', [f for f in zip(fields, test_projs)])
def test_device_forward(field, proj):
    pr = SAXSProjectorCUDA(gm)
    p = pr.forward(cuda.to_device(field))
    assert np.allclose(proj, p.copy_to_host())


@pytest.mark.parametrize('proj, adj', [f for f in zip(projs, adjs)])
def test_device_adj(proj, adj):
    pr = SAXSProjectorCUDA(gm)
    a = pr.adjoint(cuda.to_device(proj))
    print(a)
    assert np.allclose(adj, a.copy_to_host()[:, 0].ravel())


def test_str():
    dc = DataContainer('tests/test_half_circle.h5')
    pr = SAXSProjectorCUDA(dc.geometry)
    string = str(pr)
    assert '11291c' in string


def test_html():
    dc = DataContainer('tests/test_half_circle.h5')
    pr = SAXSProjectorCUDA(dc.geometry)
    html = pr._repr_html_()
    assert '11291c' in html



import logging
import hashlib
from _hashlib import HASH

import numpy as np

from numpy.typing import NDArray

logger = logging.getLogger(__name__)


def _cf_hasher(hashing_function, item) -> None:
    """ Internal method for hashing floats and complex number,
        or arrays of them. """
    if not isinstance(item, np.ndarray):
        item = np.array(item)
    item = item.ravel()
    if item.dtype.kind == 'c':
        # frexp only works on reals
        item = np.concatenate((item.real, item.imag))
    mantissa, exponent = np.frexp(item)
    edge_cases = np.isclose(abs(mantissa), 1.0, atol=1e-6, rtol=1e-6)
    mantissa[edge_cases] = np.sign(mantissa[edge_cases]) * 0.5
    exponent[edge_cases] = exponent[edge_cases] + 1
    # Round mantissa for consistency
    hashing_function.update(mantissa.round(5))
    hashing_function.update(exponent)


def _array_hasher(hashing_function: HASH, item: NDArray) -> None:
    """ Internal method for hashing arrays, lists and tuples. """
    if type(item) in (list, tuple):
        item = np.array(item)
    if item.dtype.kind in ('v'):
        hashing_function.update(item)
    # kind is bytes, int, uint, string, unicode
    if item.dtype.kind in ('biuSU'):
        hashing_function.update(np.char.encode(item.astype(str), 'utf-8'))
    # kind is float or complex
    elif item.dtype.kind in ('fc'):
        _cf_hasher(hashing_function, item)
    # unknown data, possibly ragged array etc
    else:
        raise TypeError(f'Hash of dtype `object` is not deterministic, cannot hash {item}')


def _item_hasher(hashing_function: HASH, item) -> None:
    """ Internal method for hashing floats, integers and strings. """
    if item is None:
        return
    if np.array(item).dtype.kind == 'v':
        hashing_function.update(item)
    if np.array(item).dtype.kind in ('biuSU'):
        # Cast all ints, strings, etc to string and encode
        hashing_function.update(str(item).encode('utf-8'))
    elif np.array(item).dtype.kind in ('fc'):
        _cf_hasher(hashing_function, item)
    elif np.array(item).dtype.kind == 'O':
        raise TypeError(f'Cannot hash unknown object: {item}')


def _dict_hasher(hashing_function: HASH, item) -> None:
    """ Internal method for hashing dictionaries. """
    for key, value in item.items():
        hashing_function.update(key.encode('utf-8'))
        if isinstance(value, np.ndarray) or type(value) in (list, tuple):
            _array_hasher(hashing_function, value)
        else:
            _item_hasher(hashing_function, value)


def list_to_hash(list_to_hash: list, hashing_algorithm: str = 'blake2b') -> str:
    """
    Function which takes a list containing a set of objects and automatically
    generates a deterministic hash for them.

    Parameters
    ----------
    list_to_hash
        List of a set of objects of various types, see `notes` for a complete list.
    hashing_algorithm
        The hashing algorithm to use. Can be any algorithm name in
        ``hashlib.algorithms_available``. Default is ``'blake2b'``.

    Example
    -------
    The following code snippets illustrate hashing lists that will work, and ones
    that will not work.

    Works: A list of an integer, an array, a dictionary with valid types, and a None.

    >>> from mumott.core.hashing import list_to_hash
    >>> print(list_to_hash([1, np.array((1, 3, 5)), dict(val=1, string='abc'), None]))
    2a949c...

    Does not work: an array containing a ``None``, due to the ``dtype`` being ``object``.

    >>> print(list_to_hash([np.array([None])]))
    Traceback (most recent call last):
    ...
    TypeError: Hash of dtype `object` is not deterministic, cannot hash [None]

    Does not work: a generator expression, which is an unknown object.

    >>> print(list_to_hash([(a for a in [1, 2, 3])]))
    Traceback (most recent call last):
    ...
    TypeError: Cannot hash unknown object: <generator object...

    Notes
    -----
    ``float``-type objects are rounded to five significant digits in the mantissa before hashing.
    This is necessary to obtain semi-deterministic hashes that obey a subset of fuzzy equality
    for float comparison. There are edge cases where equality can fail due to
    rounding errors, but these should be extremely rare.

    Supported entry types in :attr:`list_to_hash`:
        ``int``
            Cast to string.
            Works along with similar ``numpy`` types.
        ``float``
            Mantissa rounded to five significant digits and concatenated with exponent.
            Works along with similar ``numpy`` types.
        ``complex``
            Real and imaginary parts concatenated and treated like ``float``.
            Works along with similar ``numpy`` types.
        ``str``
            Automatically given ``'utf-8'`` encoding.
        ``bytes``
            Cast to string.
        ``None``
            Ignored.
        ``np.ndarray``
            Provided ``dtype`` is not ``object``, hence arrays of ``None`` are not allowed.
        ``list``, ``tuple``
            Provided they can be cast to allowed, i.e. non-ragged ``np.ndarray``
        ``dict``
            Assuming entries are allowed types. Keys and entries are concatenated.
            If an entry is ``None``, the key is added to the hash while the entry is ignored.
    """
    hashing_function = hashlib.new(hashing_algorithm)
    for item in list_to_hash:
        if isinstance(item, np.ndarray) or type(item) in (list, tuple):
            _array_hasher(hashing_function, item)
        elif type(item) is dict:
            _dict_hasher(hashing_function, item)
        else:
            _item_hasher(hashing_function, item)
    return hashing_function.hexdigest()



import pytest  # noqa:F401
import cupy as cp

""" This test exists to check that the CI is compatible
with `cupy`. """

a = cp.array([1, 2, 3])
b = cp.array([-0.5, -1, -4])

res = cp.array([1.5, 3, 7])

c = cp.arange(1, 10).reshape(3, 3)
d = cp.arange(1, 16).reshape(3, 5)

res2 = cp.array([[46, 52, 58, 64, 70],
                 [100, 115, 130, 145, 160],
                 [154, 178, 202, 226, 250]])


@pytest.mark.parametrize('in1, in2, result', [(a, b, res)])
def test_diff(in1, in2, result):
    out = in1 - in2
    assert cp.allclose(out, result)


@pytest.mark.parametrize('in1, in2, result', [(c, d, res2)])
def test_matmul(in1, in2, result):
    out = in1 @ in2
    assert cp.allclose(out, result)



from abc import ABC, abstractmethod
from typing import Dict

import numpy as np
from numpy.typing import NDArray

from ...core.hashing import list_to_hash


class Regularizer(ABC):

    """This is the base class from which specific regularizers are derived.
    """

    def __init__(self):
        pass

    @abstractmethod
    def get_regularization_norm(self,
                                coefficients: NDArray[float] = None,
                                get_gradient: bool = False) -> Dict:
        """Returns regularization norm and possibly gradient based on the provided coefficients.

        Parameters
        ----------
        coefficients
            An ``np.ndarray`` of values, with shape `(X, Y, Z, W)`, where
            the last channel contains e.g. tensor components.
        get_gradient
            If ``True``, returns a ``'gradient'`` of the same shape as :attr:`coefficients`.
            Otherwise, the entry ``'gradient'`` will be ``None``.

        Returns
        -------
            A dictionary with at least two entries, ``residual_norm`` and ``gradient``.
        """
        pass

    @property
    @abstractmethod
    def _function_as_str(self) -> str:
        """ Should return a string representation of the associated norm
        of the coefficients in Python idiom, e.g. 'R(x) = 0.5 * x ** 2' for L2. """
        pass

    @property
    @abstractmethod
    def _function_as_tex(self) -> str:
        """ Should return a string representation of the associated norm
        of the coefficients in MathJax-renderable TeX, e.g. $R(x) = \frac{r^2}{2}$ for L2"""
        pass

    def __hash__(self) -> int:
        to_hash = [self._function_as_str]
        return int(list_to_hash(to_hash), 16)

    def __str__(self) -> str:
        s = []
        wdt = 74
        s += ['=' * wdt]
        s += [self.__class__.__name__.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=4, precision=5, linewidth=60, edgeitems=1):
            s += ['{:18} : {}'.format('Function of coefficients', self._function_as_str)]
            s += ['{:18} : {}'.format('hash', hex(hash(self))[2:8])]
        s += ['-' * wdt]
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += [f'<h3>{__class__.__name__}</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>']
        s += ['<tbody>']
        with np.printoptions(threshold=4, edgeitems=2, precision=2, linewidth=40):
            s += ['<tr><td style="text-align: left;">Function of coefficients</td>']
            s += [f'<td>1</td><td>{self._function_as_tex}</td></tr>']
            s += ['<tr><td style="text-align: left;">Hash</td>']
            h = hex(hash(self))
            s += [f'<td>{len(h)}</td><td>{h[2:8]}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        return '\n'.join(s)



import pytest
import numpy as np
from mumott.core.spherical_harmonic_mapper import SphericalHarmonicMapper


@pytest.mark.parametrize('test_input,outcome',
                         [
                             (dict(max_order=4,
                                   polar_resolution=7,
                                   azimuthal_resolution=4,
                                   polar_zero=0,
                                   azimuthal_zero=0,
                                   ), None),
                             (dict(max_order=8,
                                   polar_resolution=13,
                                   azimuthal_resolution=21,
                                   polar_zero=np.pi / 3,
                                   azimuthal_zero=np.pi / 8,
                                   ), None),
                             (dict(max_order=6,
                                   polar_resolution=128,
                                   azimuthal_resolution=256,
                                   polar_zero=0,
                                   azimuthal_zero=0,
                                   ), None),
                         ])
def test_init(test_input, outcome):
    if outcome is None:
        S = SphericalHarmonicMapper(**test_input)
        assert S.max_order == test_input['max_order']
        assert np.all(S.orders == [i for i in range(0, test_input['max_order'] + 1, 2)
                                   for j in range(2 * i + 1)])
        assert np.all(S.degrees == [j for i in range(0, test_input['max_order'] + 1, 2)
                                    for j in range(-i, i + 1)])
        assert np.all(S.phi.shape == (test_input['polar_resolution'], test_input['azimuthal_resolution']))
        assert np.all(S.theta.shape == (test_input['polar_resolution'], test_input['azimuthal_resolution']))
        assert S.polar_zero == test_input['polar_zero']
        assert S.azimuthal_zero == test_input['azimuthal_zero']
        assert np.all(S.theta_interpolator.shape ==
                      (test_input['polar_resolution'], test_input['azimuthal_resolution'] - 1))
        assert np.all(S.phi_interpolator.shape ==
                      (test_input['polar_resolution'], test_input['azimuthal_resolution'] - 1))
        assert np.all(S.map.shape == (test_input['polar_resolution'],
                                      test_input['azimuthal_resolution'],
                                      (test_input['max_order'] // 2 + 1) * (test_input['max_order'] + 1)))
        if (test_input['azimuthal_resolution'] == 128) & \
           (test_input['polar_resolution'] == 256) & (test_input['max_order'] == 6):
            coefficients = np.array([-0.6853447, 1.79851744, -1.23394354, -0.20025308, -2.01433032,
                                     -0.63057234, -0.68189296, 0.09721236, 1.25074053, 0.02795743,
                                     0.9053322, 0.75103781, -3.11883902, 1.96487535, -1.16443111,
                                     -0.06631476, -0.67329735, -0.48743056, 0.09240111, 1.36270693,
                                     0.84001768, -2.31092023, -0.21464363, 1.04476394, -0.77695792,
                                     -1.58423597, 0.87403145, 0.70022989])
            amplitudes = S.get_amplitudes(coefficients)
            fit = S.get_harmonic_coefficients(amplitudes[:, :-1])
            expected = np.array([-0.67149577, 1.80497953, -1.23877695, -0.18810264, -2.02513128,
                                 -0.61917466, -0.68518089, 0.09803488, 1.253694, 0.02785181,
                                 0.8857664, 0.76285412, -3.09020449, 1.9667663, -1.16317915,
                                 -0.06656277, -0.67505851, -0.49411828, 0.09223364, 1.36925961,
                                 0.84341165, -2.30343654, -0.21779937, 1.02573042, -0.77279774,
                                 -1.54228925, 0.87009572, 0.70290764])
            assert np.allclose(fit, expected)
    else:
        with pytest.raises(outcome):
            _ = SphericalHarmonicMapper(**test_input)


def test_setter():
    S = SphericalHarmonicMapper(max_order=2)
    S.max_order = 6
    assert S.max_order == 6
    assert S.orders.size == 28
    assert np.all(S.orders == [i for i in range(0, 7, 2) for j in range(2 * i + 1)])
    assert np.all(S.degrees == [j for i in range(0, 7, 2)
                                for j in range(-i, i + 1)])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import sys
from setuptools import find_packages, setup


if sys.version_info < (3, 8, 0, 'final', 0):
    raise SystemExit('Python 3.8 or later is required!')

# Fetch information available in module
with open('README.rst', encoding='utf-8') as fd:
    long_description = fd.read()

with open('mumott/__init__.py', encoding='utf-8') as fd:
    lines = '\n'.join(fd.readlines())

version = re.search("__version__ = '(.*)'", lines).group(1)
description = re.search("__description__ = '(.*)'", lines).group(1)
maintainer = re.search("__maintainer__ = '(.*)'", lines).group(1)
url = re.search("__url__ = '(.*)'", lines).group(1)
license = re.search("__license__ = '(.*)'", lines).group(1)

# Set classifiers used on PyPI
classifiers = [
    'Development Status :: 4 - Beta',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: {}'.format(license),
    'Topic :: Scientific/Engineering :: Chemistry',
    'Topic :: Scientific/Engineering :: Image Processing',
    'Topic :: Scientific/Engineering :: Physics',
    'Topic :: Scientific/Engineering :: Visualization']

# Python libraries this package depends on
python_dependencies = ['colorcet',
                       'colorspacious',
                       'matplotlib',
                       'h5py',
                       'numba<=0.56.3',
                       'numpy',
                       'scipy']

# Carry out the actual package installation and setup
setup(name='mumott',
      py_modules=['mumott'],
      version=version,
      description=description,
      long_description=long_description,
      maintainer=maintainer,
      url=url,
      license=license,
      classifiers=classifiers,
      install_requires=python_dependencies,
      packages=find_packages(),
      zip_safe=False)



""" Container for the class ProjectionViewer. """

import logging

import numpy as np
import matplotlib.pyplot as plt
import colorcet  # noqa

from colorspacious import cspace_converter
from matplotlib import cm
from matplotlib.colors import Colormap
from matplotlib.gridspec import GridSpec
from numpy.typing import NDArray
from typing import Union, Tuple

from .projection_frame import ProjectionFrame
from .orientation_image_mapper import OrientationImageMapper
from mumott.data_handling import DataContainer


class ProjectionViewer:
    """
    Class for viewing data and synthetic projections in a scrollable plot.
    The viewed projection can be changed with either the arrow keys or
    with the mouse scrolling wheel.

    Parameters
    ----------
    data_container : DataContainer
        A :class:`DataContainer
        <mumott.core.data_container.DataContainer>` object, containing
        the necessary dimension and rotation information to show orientations correctly.
    data : numpy.ndarray(dtype=numpy.float64), optional
        A flat array of data from e.g. reconstruction output. If not given, data
        will be loaded from the provided ``data_container``. Provide if you e.g.
        want to look at synthetic projections from a reconstruction.
    orientation_symmetry : ``'longitudinal'``, ``'transversal'``, optional
        Specifies the assumed three-dimensional symmetry of the reciprocal
        space map, which determines whether the minimum or the maximum of the
        ``cos(phi) ** 2`` component of the of the intensity is interpreted as
        the orientation angle. The default is ``'longitudinal'``, which interprets the
        angle of maximum scattering as the orientation, corresponding to
        an assumed three-dimensional symmetry of polar caps of intensity.
        The other option is ``transversal``, which corresponds to a symmetry
        of a great circle of intensity (also known as fiber symmetry), corresponding
        to a phase shift of ``90`` degrees.
    mean_colormap : str or matplotlib.colors.Colormap
        Specifies the colormap of the mean intensity across all angles. The default
        is ``'cet_linear_bmy_10_95_c71'``, a linear high-contrast with blue, magenta
        and yellow.
    std_colormap : str or matplotlib.colors.Colormap
        Specifies the colormap of the standar deviation of the data across all angles.
        The default is ``'cet_gouldian'``, a linear high-contrast map with blue,
        green and yellow.
    phase_colormap : str or matplotlib.colors.Colormap
        Specifies the colormap of the phase of the ``cos(detector_angles) ** 2`` fit
        of the data. The default is ``'cet_CET_C10'``, an isoluminant cyclic map.
    mean_range : tuple(float, float)
        Specifies the range of the colormap of the mean.
    std_range : tuple(float, float)
        Specifies the range of the colormap of the standard deviation.
    """

    def __init__(self,
                 data_container: DataContainer,
                 data: NDArray[np.float64] = None,
                 orientation_symmetry: str = 'longitudinal',
                 mean_colormap: Union[str, Colormap] = 'cet_linear_bmy_10_95_c71',
                 std_colormap: Union[str, Colormap] = 'cet_gouldian',
                 phase_colormap: Union[str, Colormap] = 'cet_CET_C10',
                 mean_range: Tuple[float, float] = None,
                 std_range: Tuple[float, float] = None):
        detector_angles = data_container.geometry.detector_angles
        det_vector = \
            (np.cos(detector_angles[np.newaxis, :]) *
             data_container.geometry.detector_direction_origin[:, np.newaxis] +
             np.sin(detector_angles[np.newaxis, :]) *
             data_container.geometry.detector_direction_positive_90[:, np.newaxis])
        distance_j = np.dot(data_container.geometry.j_direction_0[np.newaxis, :],
                            det_vector).squeeze()
        distance_k = np.dot(data_container.geometry.k_direction_0[np.newaxis, :],
                            det_vector).squeeze()
        detector_angles = np.arctan2(distance_k, distance_j)
        if data is None:
            data = data_container.data
        if not np.allclose(np.diff(detector_angles).reshape(1, -1),
                           np.diff(detector_angles).reshape(-1, 1),
                           rtol=2e-01, atol=8e-02):
            logging.warning('The provided `detector_angles` do not appear to be '
                            'sorted and equally spaced to within an absolute tolerance '
                            'of 5 degrees and a relative tolerance of twenty percent of the '
                            'segment width.'
                            '\nThe orientation angle may not behave as expected. Please be cautious'
                            'in interpreting the figure.')
        if orientation_symmetry == 'transversal':
            detector_angles = np.arctan2(np.sin(detector_angles), np.cos(detector_angles))
        elif orientation_symmetry == 'longitudinal':
            detector_angles = np.arctan2(-np.cos(detector_angles), np.sin(detector_angles))
        else:
            logging.warning(
                    'Unknown `orientation_symmetry` option: "' + orientation_symmetry +
                    '"\nUsing default value: "longitudinal".\n'
                    'Valid options are "transversal" and "longitudinal".')
            detector_angles = np.arctan2(-np.cos(detector_angles), np.sin(detector_angles))
        direction = np.arctan2(np.sin(detector_angles[1] - detector_angles[0]),
                               np.cos(detector_angles[1] - detector_angles[0]))
        a = float(np.round(direction >= 0))
        b = np.exp((-2j) * detector_angles[0])
        self._mean_colormap = mean_colormap
        self._std_colormap = std_colormap
        self._phase_colormap = phase_colormap
        self._cielab_converter = cspace_converter('sRGB255', 'JCh')
        self._srgb_converter = cspace_converter('JCh', 'sRGB255')
        self._orientation_image_mapper = OrientationImageMapper(phase_colormap)
        self._wheel_properties = self._orientation_image_mapper.wheel_properties
        self._phase_colormap_lut = cm.get_cmap(self._phase_colormap, 256)
        self._projection_number = len(data)
        self._clims_mean = np.percentile(
            a=data.mean(axis=2),
            q=(1., 99.9))
        if mean_range is not None:
            if mean_range[0] is not None:
                self._clims_mean[0] = mean_range[0]
            if mean_range[1] is not None:
                self._clims_mean[1] = mean_range[1]
        self._clims_std = np.percentile(
            a=data.std(axis=2),
            q=(1., 99.9))
        if std_range is not None:
            if std_range[0] is not None:
                self._clims_std[0] = std_range[0]
            if std_range[1] is not None:
                self._clims_std[1] = std_range[1]
        self._projections = []
        for i, frame in enumerate(data):
            temp_mean = frame[:, :, :detector_angles.size].mean(axis=2).T
            temp_std = frame[:, :, :detector_angles.size].std(axis=2).T
            temp_fft = np.fft.rfft(frame[:, :, :detector_angles.size] ** 2, axis=2, norm='ortho')
            if a:
                temp_angle = (np.sqrt(temp_fft[:, :, 1] * b)).T
            else:
                temp_angle = (np.sqrt(np.conj(temp_fft[:, :, 1]) * b)).T
            temp_angle = np.angle(np.sign(temp_angle.imag) * temp_angle)
            temp_mask = np.round(
                np.sqrt(
                    abs(temp_fft[:, :, 0])) > 0.5 * np.sqrt(abs(temp_fft[:, :, 0])).mean()).astype(float).T
            temp_aniso_factor = np.sqrt(
                abs(temp_fft[:, :, 1]) / (np.finfo(float).tiny + abs(temp_fft[:, :, 0]))).T
            temp_shader = temp_aniso_factor.clip(0., 1)
            phase_rgba = self._phase_colormap_lut(temp_angle / np.pi)
            phase_cielab = self._cielab_converter(phase_rgba[:, :, :-1])
            phase_cielab[:, :, :2] *= 1.1 * ((1 - np.exp(-temp_shader)) /
                                             (1 - np.exp(-1))).reshape(temp_shader.shape + (1,))
            phase_rgba[:, :, :-1] = self._srgb_converter(phase_cielab).clip(0., 1.)
            phase_rgba[:, :, -1] = temp_mask
            temp_frame = ProjectionFrame(
                mean=temp_mean,
                std=temp_std,
                angle_color_quad=phase_rgba)
            self._projections.append(temp_frame)
        self._fig = plt.figure()
        self._grid_spec = GridSpec(6, 9, figure=self._fig)
        self._axes = []
        self._axes.append(self._fig.add_subplot(self._grid_spec[0:4, 0:3]))
        self._axes.append(self._fig.add_subplot(self._grid_spec[0:4, 3:6]))
        self._axes.append(self._fig.add_subplot(self._grid_spec[0:4, 6:9]))
        self._axes.append(self._fig.add_subplot(self._grid_spec[4:6, 6:9], projection='polar'))
        self._colorbar_axes = []
        self._colorbar_axes.append(self._fig.add_subplot(self._grid_spec[5, 0:3]))
        self._colorbar_axes.append(self._fig.add_subplot(self._grid_spec[5, 3:6]))
        for a in self._axes:
            a.grid(False)
            a.set_yticks([])
            a.set_xticks([])
            a.set_yticklabels([])
            a.set_xticklabels([])
        self._im1 = self._axes[0].imshow(
            self._projections[0].mean, origin='lower', cmap=self._mean_colormap)
        self._axes[0].set_title('Mean')
        self._fig.colorbar(
            self._im1, cax=self._colorbar_axes[0], orientation='horizontal')
        self._im2 = self._axes[1].imshow(
            self._projections[0].std, origin='lower', cmap=self._std_colormap)
        self._axes[1].set_title('Standard deviation')
        self._fig.colorbar(self._im2, cax=self._colorbar_axes[1], orientation='horizontal')

        self._im3 = self._axes[2].imshow(self._projections[0].angle_color_quad,
                                         origin='lower',)
        self._axes[2].set_title('Orientation phase')
        self._axes[2].set_facecolor((254 / 255, 254 / 255, 254 / 255))
        self._wheel_rgba = self._wheel_properties[2]
        self._axes[3].set_facecolor((254 / 255, 254 / 255, 254 / 255))
        wheel_shader = self._wheel_properties[1][:-1, :-1] ** 2
        wheel_cielab = self._cielab_converter(self._wheel_rgba[:, :, :-1])
        wheel_cielab[:, :, :2] *= 1.1 * ((1 - np.exp(-wheel_shader)) /
                                         (1 - np.exp(-1))).reshape(wheel_shader.shape + (1,))
        wheel_rgb = self._srgb_converter(wheel_cielab).clip(0., 1.)
        self._wheel_rgba[:, :, :-1] = wheel_rgb
        self._axes[3].grid(False)
        self._color_wheel = self._axes[3].pcolormesh(self._wheel_properties[0],
                                                     1 - np.arccos(self._wheel_properties[1]) * 2 / np.pi,
                                                     wheel_shader[:, :],
                                                     facecolors=self._wheel_rgba.reshape(-1, 4),
                                                     edgecolors=self._wheel_rgba.reshape(-1, 4))
        self._axes[3].set_rmax(1.)
        self._axes[3].set_rmin(0.)
        self._projection_ind = 0

    def change_frame(self, ind: int) -> None:
        """
        Method for updating the projection shown
        when the projection slider changes.

        Parameters
        ----------
        ind
            The index of the projection to be shown.
        """
        if ind != self._projection_ind:
            self._projection_ind = ind
            ind = int(ind)
            self._im1.set_data(self._projections[ind].mean)
            self._im1.set_clim(*self._clims_mean)
            self._im2.set_data(self._projections[ind].std)
            self._im2.set_clim(*self._clims_std)
            self._im3.set_data(self._projections[ind].angle_color_quad)
            self._im3.autoscale()
            self._axes[0].relim()
            self._axes[1].relim()
            self._axes[2].relim()
            self._axes[0].autoscale_view()
            self._axes[1].autoscale_view()
            self._axes[2].autoscale_view()
            self._axes[0].set_aspect('equal')
            self._axes[1].set_aspect('equal')
            self._axes[2].set_aspect('equal')
            self._fig.canvas.draw_idle()
            self._fig.canvas.flush_events()



import logging
import os
from multiprocessing import cpu_count
import numba


logger = logging.getLogger(__name__)


def numba_setup() -> None:
    """This function sets defaults for environmental variables and numba
    parameters to achieve a balanced performance for most situations.
    Advanced users may want to customize these parameters.
    """

    if cpu_count() >= 6:
        number_of_threads = 6
    else:
        number_of_threads = cpu_count()
        os.environ['NUMBA_NUM_THREADS'] = f'{number_of_threads}'
        os.environ['OMP_NUM_THREADS'] = f'{number_of_threads}'
    logger.info(f'Setting the number of threads to {number_of_threads}')

    if numba.config.NUMBA_DEFAULT_NUM_THREADS >= number_of_threads:
        numba.config.NUMBA_DEFAULT_NUM_THREADS = number_of_threads
        numba.config.NUMBA_NUM_THREADS = number_of_threads
        numba.set_num_threads(number_of_threads)

    # numba.cuda has a very high output of `INFO`-level messages.
    numba_logger = logging.getLogger('numba')
    numba_logger.setLevel(logging.WARNING)
    logger.info('Setting numba log level to WARNING.')



import pytest
from mumott.output_handling.saving import dict_to_h5

import numpy as np
import h5py


class myclass:
    a = 0


list_of_dicts = [dict(a=1, b=2, c=3),
                 dict(arr=np.array([0]), inner=dict(a=1, b=(1, 2, 3))),
                 dict(myst='stringstring', bst=b'bytestring', arr=np.array(['array', 'of', 'strings'])),
                 dict(co=1 + 1j, arr=np.array((1 + 1j)), ll=[np.array((1, 2)), np.array((4, 5))]),
                 dict(si=np.sin(np.pi / 5), sq=np.sqrt(2) / 10, sqrt=np.sqrt(3) * 1e-32),
                 dict(nope=None, dinope=dict(val=None)),
                 dict(arr=np.array([None])),
                 dict(myc=myclass())]

list_of_bools = [True, True, True, True, True, False, False, False]


def recursive_read(items, folder):
    for key, val in items.items():
        if isinstance(val, dict):
            recursive_read(val, folder[key])
        else:
            print(key, val, folder[key])
            # this monstrosity brought to you by encoding
            if np.array(val).dtype.kind in 'SU':
                if np.array(val).dtype.kind == 'S':
                    val = val.decode()
                if folder[key].size > 1:
                    for i in range(folder[key].size):
                        assert val[i] == folder[key][i].decode()
                else:
                    assert np.all(val == folder[key][0].decode())
            else:
                assert np.allclose(val, folder[key])


@pytest.mark.parametrize('in_dict, list_bool', [t for t in zip(list_of_dicts, list_of_bools)])
def test_saving(in_dict, list_bool, tmp_path):
    d = tmp_path / "sub"
    d.mkdir()
    p = d / "hello.h5"
    if list_bool is False:
        dict_to_h5(in_dict, p, True)
    else:
        dict_to_h5(in_dict, p, True)
        with h5py.File(p) as file:
            recursive_read(in_dict, file)



import logging

from typing import Tuple

import numpy as np
from numpy.typing import NDArray
from numba import get_num_threads


from ...core.john_transform import john_transform, john_transform_adjoint
from ...core.hashing import list_to_hash
from ... import Geometry
from .base_projector import Projector

logger = logging.getLogger(__name__)


class SAXSProjectorNumba(Projector):
    """
    Projector for transforms of tensor fields from three-dimensional space
    to projection space.

    Parameters
    ----------
    geometry : Geometry
        An instance of :class:`Geometry <mumott.Geometry>` containing the
        necessary vectors to compute forwared and adjoint projections.
    step_size : np.float64
        A number ``0 < step_size <= 1`` that indicates the degree of
        upsampling when computing projections. Default is ``0.5``.
    """
    def __init__(self,
                 geometry: Geometry,
                 step_size: np.float64 = np.float64(0.5)):

        super().__init__(geometry)

        self._sampling_kernel = np.array((1.,), dtype=np.float64)
        self._kernel_offsets = np.array((0., 0.), dtype=np.float64)
        self._created_sampling_kernel = False
        self.step_size = step_size
        self._basis_vector_projection = np.zeros((len(self._geometry), 3), dtype=np.float64)
        self._basis_vector_j = np.zeros((len(self._geometry), 3), dtype=np.float64)
        self._basis_vector_k = np.zeros((len(self._geometry), 3), dtype=np.float64)
        self._update(force_update=True)

    def _get_john_transform_parameters(self,
                                       index: int) -> Tuple:
        vector_p = self._basis_vector_projection[index]
        vector_j = self._basis_vector_j[index]
        vector_k = self._basis_vector_k[index]
        projection_offsets = np.array((self._geometry[index].j_offset,
                                       self._geometry[index].k_offset)).ravel()
        return (vector_p, vector_j, vector_k, self._step_size, projection_offsets,
                self._sampling_kernel.ravel(), self._kernel_offsets.reshape(2, -1))

    def forward(self,
                field: NDArray,
                indices: NDArray[int] = None) -> NDArray:
        """ Compute the forward projection of a tensor field.

        Parameters
        ----------
        field
            An array containing coefficients in its fourth dimension,
            which are to be projected into two dimensions. The first three
            dimensions should match the ``volume_shape`` of the sample.
        indices
            A one-dimensional array containing one or more indices
            indicating which projections are to be computed. If ``None``,
            all projections will be computed.

        Returns
        -------
            An array with four dimensions ``(I, J, K, L)``, where
            the first dimension matches :attr:`indices`, such that
            ``projection[i]`` corresponds to the geometry of projection
            ``indices[i]``. The second and third dimension contain
            the pixels in the ``J`` and ``K`` dimension respectively, whereas
            the last dimension is the coefficient dimension, matching ``field[-1]``.
        """
        assert np.allclose(field.shape[:-1], self._geometry.volume_shape)
        self._update()
        if indices is None:
            return self._forward_stack(field)
        return self._forward_subset(field, indices)

    def _forward_subset(self,
                        field: NDArray,
                        indices: NDArray[int]) -> NDArray:
        """ Internal method for computing a subset of projections.

        Parameters
        ----------
        field
            The field to be projected.
        indices
            The indices indicating the subset of all projections in the
            system geometry to be computed.

        Returns
        -------
            The resulting projections.
        """
        indices = np.array(indices).ravel()
        projections = np.zeros((indices.size,) +
                               tuple(self._geometry.projection_shape) +
                               (field.shape[-1],), dtype=np.float64)
        self._check_indices_kind_is_integer(indices)
        for i, index in enumerate(indices):
            john_transform(projections[i], field, *self._get_john_transform_parameters(index))
        return projections

    def _forward_stack(self,
                       field: NDArray) -> NDArray:
        """Internal method for forward projecting an entire stack.

        Parameters
        ----------
        field
            The field to be projected.

        Returns
        -------
            The resulting projections.
        """
        projections = np.zeros((len(self._geometry),) +
                               tuple(self._geometry.projection_shape) +
                               (field.shape[-1],), dtype=np.float64)
        for i, g in enumerate(self._geometry):
            john_transform(projections[i], field, *self._get_john_transform_parameters(i))
        return projections

    def adjoint(self,
                projections: NDArray,
                indices: NDArray[int] = None) -> NDArray:
        """ Compute the adjoint of a set of projections according to the system geometry.

        Parameters
        ----------
        projections
            An array containing coefficients in its last dimension,
            from e.g. the residual of measured data and forward projections.
            The first dimension should match :attr:`indices` in size, and the
            second and third dimensions should match the system projection geometry.
        indices
            A one-dimensional array containing one or more indices
            indicating from which projections the adjoint is to be computed.

        Returns
        -------
            The adjoint of the provided projections.
            An array with four dimensions ``(X, Y, Z, P)``, where the first
            three dimensions are spatial and the last dimension runs over
            coefficients.
        """
        assert np.allclose(projections.shape[-3:-1], self._geometry.projection_shape)
        self._update()
        if indices is None:
            return self._adjoint_stack(projections)
        return self._adjoint_subset(projections, indices)

    def _adjoint_subset(self,
                        projections: NDArray,
                        indices: NDArray[int]) -> NDArray:
        """ Internal method for computing the adjoint of only a subset of projections.

        Parameters
        ----------
        projections
            An array containing coefficients in its last dimension,
            from e.g. the residual of measured data and forward projections.
            The first dimension should match :attr:`indices` in size, and the
            second and third dimensions should match the system projection geometry.
        indices
            A one-dimensional array containing one or more indices
            indicating from which projections the adjoint is to be computed.

        Returns
        -------
            The adjoint of the provided projections.
            An array with four dimensions ``(X, Y, Z, P)``, where the first
            three dimensions are spatial and the last dimension runs over
            coefficients. """
        if projections.ndim == 3:
            assert indices.size == 1
            projections = projections[np.newaxis, ...]
        else:
            assert indices.size == projections.shape[0]
        indices = np.array(indices).ravel()
        self._check_indices_kind_is_integer(indices)
        field = np.zeros((get_num_threads(),) +
                         tuple(self._geometry.volume_shape) +
                         (projections.shape[-1],), dtype=np.float64)
        for num, i in enumerate(indices):
            # Note that we assume projections match ``indices`` in their layout.
            john_transform_adjoint(projections[num], field, *self._get_john_transform_parameters(i))
        # each thread writes into a separate array, they are reduced by summation.
        return field.sum(0)

    def _adjoint_stack(self,
                       projections: NDArray) -> NDArray:
        """ Internal method for computing the adjoint of a whole stack of projections.

        Parameters
        ----------
        projections
            An array containing coefficients in its last dimension,
            from e.g. the residual of measured data and forward projections.
            The first dimension should run over all the projection directions
            in the system geometry.

        Returns
        -------
            The adjoint of the provided projections.
            An array with four dimensions ``(X, Y, Z, P)``, where the first
            three dimensions are spatial, and the last dimension runs over
            coefficients. """
        assert projections.shape[0] == len(self._geometry)
        field = np.zeros((get_num_threads(),) +
                         tuple(self._geometry.volume_shape) +
                         (projections.shape[-1],), dtype=np.float64)
        for i, g in enumerate(self._geometry):
            john_transform_adjoint(projections[i], field, *self._get_john_transform_parameters(i))
        # each thread writes into a separate array, they are reduced by summation.
        return field.sum(0)

    def _update(self,
                force_update: bool = False) -> None:
        super()._update(force_update)
        self._calculate_basis_vectors()
        self._geometry_hash = hash(self._geometry)

    def _calculate_basis_vectors(self) -> None:
        """ Calculates the basis vectors for the John transform, one projection vector
        and two coordinate vectors. """
        self._basis_vector_projection = np.einsum(
            'kij,i->kj', self._geometry.rotations_as_array, self._geometry.p_direction_0)
        self._basis_vector_j = np.einsum(
            'kij,i->kj', self._geometry.rotations_as_array, self._geometry.j_direction_0)
        self._basis_vector_k = np.einsum(
            'kij,i->kj', self._geometry.rotations_as_array, self._geometry.k_direction_0)

    def create_sampling_kernel(self,
                               kernel_dimensions: Tuple[int, int] = (3, 3),
                               kernel_width: Tuple[float, float] = (1., 1.),
                               kernel_type: str = 'bessel') -> None:
        """ Creates a kernel to emulate the point spread function (PSF) of the
        beam. This can improve the accuracy of the projection function.

        Parameters
        ----------
        kernel_dimensions
            A tuple of how many points should be sampled in each direction of the kernel.
            The total number of line integrals per pixel in the data is
            ``kernel_dimensions[0] * kernel_dimensions[1]``.
        kernel_width
            Width parameter for the kernel in units of pixels.
            Typically the full-width-half-maximum (FWHM) of the beam used for data acquisition.
        kernel_type
            The type of kernel to use.
            Acceptable values are ``'bessel'``, ``'rectangular'``, and ``'gaussian'``.

            ``'bessel'`` uses a sinc function multiplied by a
            Lanczos window with the width parameter being the first Bessel zero. This
            gives a sharply peaked distribution that goes to zero at twice the full
            width half maximum, and samples are taken up to this zero.

            ``'rectangular'`` samples uniformly in a rectangle of size :attr:`kernel_width`.

            ``'gaussian'`` uses a normal distribution with the FWHM given by
            :attr:`kernel_width` sampled up to twice the FWHM.
        """
        if self._created_sampling_kernel:
            logger.warning('It appears that you have already created a sampling kernel.'
                           ' The old sampling kernel will be overwritten.')
        if kernel_type == 'bessel':
            L = 0.7478
            ji = np.linspace(-kernel_width[0], kernel_width[0], kernel_dimensions[0] * 100)
            fj = np.sinc(ji * L / kernel_width[0]) * np.sinc(ji / kernel_width[0])
            ki = np.linspace(-kernel_width[1], kernel_width[1], kernel_dimensions[1] * 100)
            fk = np.sinc(ki * L / kernel_width[1]) * np.sinc(ki / kernel_width[1])
        elif kernel_type == 'rectangular':
            ji = np.linspace(-kernel_width[0] / 2, kernel_width[0] / 2, kernel_dimensions[0] * 100)
            fj = np.ones_like(ji)
            ki = np.linspace(-kernel_width[1] / 2, kernel_width[1] / 2, kernel_dimensions[1] * 100)
            fk = np.ones_like(ki)
        elif kernel_type == 'gaussian':
            std = np.array(kernel_width) / (2 * np.sqrt(2 * np.log(2)))
            ji = np.linspace(-kernel_width[0], kernel_width[0], kernel_dimensions[0] * 100)
            fj = np.exp(-0.5 * (ji ** 2) / ((std[0]) ** 2))
            ki = np.linspace(-kernel_width[1], kernel_width[1], kernel_dimensions[1] * 100)
            fk = np.exp(-0.5 * (ki ** 2) / ((std[1]) ** 2))
        else:
            raise ValueError(f'Unknown kernel type: {kernel_type}.')
        fr = fj.reshape(-1, 1) * fk.reshape(1, -1)
        fr = fr.reshape(kernel_dimensions[0], 100, kernel_dimensions[1], 100)
        fi = fr.sum(axis=(1, 3))
        fi = (fi / fi.sum()).astype(np.float64)
        J = ji.reshape(-1, 1) * np.ones((1, ki.size))
        K = ki.reshape(1, -1) * np.ones((ji.size, 1))
        J = J.reshape(kernel_dimensions[0], 100, kernel_dimensions[1], 100)
        K = K.reshape(kernel_dimensions[0], 100, kernel_dimensions[1], 100)
        Ji = J.mean(axis=(1, 3)).flatten()
        Ki = K.mean(axis=(1, 3)).flatten()
        offset = np.concatenate((Ji, Ki)).astype(np.float64)
        self._sampling_kernel = fi
        self._kernel_offsets = offset
        self._sampling_kernel_size = np.int32(kernel_dimensions[0] * kernel_dimensions[1])
        self._created_sampling_kernel = True

    @property
    def step_size(self) -> np.float64:
        """
        One-dimensional upsampling ratio for the integration of each projection line.
        """
        return self._step_size

    @step_size.setter
    def step_size(self,
                  val: np.float64) -> None:
        if not 0.0 < val <= 1.0:
            raise ValueError('Integration step size must be less than or'
                             ' equal to 1.0, and greater than 0.0, but the'
                             f' provided value is {val}.')
        # under normal circumstances this casting is not needed, but it is a safeguard
        self._step_size = np.float64(val)

    @property
    def sampling_kernel(self) -> NDArray[np.float64]:
        """ Convolution kernel for the sampling profile. """
        return self._sampling_kernel

    @property
    def dtype(self) -> np.typing.DTypeLike:
        """ Preferred dtype of this ``Projector``. """
        return np.float64

    @property
    def kernel_offsets(self) -> NDArray[np.float64]:
        """ Offsets for the sampling profile kernel. """
        return self._kernel_offsets

    def __hash__(self) -> int:
        to_hash = [self._step_size,
                   self._sampling_kernel,
                   self._kernel_offsets,
                   self._basis_vector_projection,
                   self._basis_vector_j,
                   self._basis_vector_k,
                   self._geometry_hash,
                   hash(self._geometry)]
        return int(list_to_hash(to_hash), 16)

    def __str__(self) -> str:
        wdt = 74
        s = []
        s += ['-' * wdt]
        s += [self.__class__.__name__.center(wdt)]
        s += ['-' * wdt]
        with np.printoptions(threshold=4, edgeitems=2, precision=5, linewidth=60):
            s += ['{:18} : {}'.format('step_size', self.step_size)]
            s += ['{:18} : {}'.format('is_dirty', self.is_dirty)]
            s += ['{:18} : {}'.format('sampling_kernel', self.sampling_kernel)]
            s += ['{:18} : {}'.format('kernel_offsets', self.kernel_offsets)]
            s += ['{:18} : {}'.format('hash', hex(hash(self))[2:8])]
        s += ['-' * wdt]
        return '\n'.join(s)

    def _repr_html_(self) -> str:
        s = []
        s += [f'<h3>{self.__class__.__name__}</h3>']
        s += ['<table border="1" class="dataframe">']
        s += ['<thead><tr><th style="text-align: left;">Field</th><th>Size</th><th>Data</th></tr></thead>']
        s += ['<tbody>']
        with np.printoptions(threshold=4, edgeitems=2, precision=2, linewidth=40):
            s += ['<tr><td style="text-align: left;">step_size</td>']
            s += [f'<td>1</td><td>{self.step_size}</td></tr>']
            s += ['<tr><td style="text-align: left;">is_dirty</td>']
            s += [f'<td>1</td><td>{self.is_dirty}</td></tr>']
            s += ['<tr><td style="text-align: left;">sampling_kernel</td>']
            s += [f'<td>{self.sampling_kernel.shape}</td><td>{self.sampling_kernel}</td></tr>']
            s += ['<tr><td style="text-align: left;">kernel_offsets</td>']
            s += [f'<td>{self.kernel_offsets.shape}</td><td>{self.kernel_offsets}</td></tr>']
            s += ['<tr><td style="text-align: left;">hash</td>']
            s += [f'<td>{len(hex(hash(self)))}</td><td>{hex(hash(self))[2:8]}</td></tr>']
        s += ['</tbody>']
        s += ['</table>']
        return '\n'.join(s)
